#!/usr/bin/env python
"""
This module contains the class Pyqtgrah which is used to handle the histogram widget in the GUI.
"""

from ibeatles import DataType


class Pyqtgrah:
    list_steps_without_combine_algo_options = (
        DataType.normalization,
        DataType.bin,
        DataType.fitting,
    )

    def __init__(
        self,
        parent=None,
        image_view=None,
        data_type=DataType.sample,
        add_mean_radio_button_changed=False,
    ):
        self.parent = parent
        self.image_view = image_view
        self.data_type = data_type
        self.first_update = False
        self.histo_widget = self.image_view.getHistogramWidget()
        self.add_mean_radio_button_changed = add_mean_radio_button_changed

        if self.data_type == DataType.normalized:
            self.combine_algo = "sum" if self.parent.ui.roi_add_button.isChecked() else "mean"
        elif self.data_type == DataType.sample:
            self.combine_algo = "sum" if self.parent.ui.roi_mean_button.isChecked() else "mean"
        else:
            self.combine_algo = "sum"

    def set_state(self, state=None):
        if self.parent.image_view_settings[self.data_type]["first_time_using_state"]:
            self.parent.image_view_settings[self.data_type]["first_time_using_state"] = False
            return

        _view = self.image_view.getView()
        _view_box = _view.getViewBox()
        if not state:
            state = self.parent.image_view_settings[self.data_type]["state"]
            if not state:
                return
        _view_box.setState(state)

    def get_state(self):
        _view = self.image_view.getView()
        _view_box = _view.getViewBox()
        _state = _view_box.getState()
        return _state

    def save_histogram_level(self, data_type_of_data=None):
        """we are going to save the histogram but only if we didn't come here via the add/mean radio buttons"""

        # we switched the 'mean' and 'add' buttons, no need to save the histogram then
        if self.add_mean_radio_button_changed:
            return

        if self.data_type in self.list_steps_without_combine_algo_options:
            if self.parent.image_view_settings[self.data_type]["first_time_using_histogram"]:
                # this is to make bin and fit tabs working
                self.parent.image_view_settings[self.data_type]["first_time_using_histogram"] = False
                return
        else:
            if self.parent.image_view_settings[self.data_type]["first_time_using_histogram"][self.combine_algo]:
                # this is to make bin and fit tabs working
                self.parent.image_view_settings[self.data_type]["first_time_using_histogram"][self.combine_algo] = False
                return

        if data_type_of_data is None:
            data_type_of_data = self.data_type

        if len(self.parent.data_metadata[data_type_of_data]["data"]) == 0:
            return

        if self.data_type == DataType.normalization:
            histogram_level = self.parent.image_view_settings[data_type_of_data]["histogram"]
        else:
            histogram_level = self.parent.image_view_settings[data_type_of_data]["histogram"][self.combine_algo]

        if histogram_level is None:
            self.first_update = True
        _histo_widget = self.image_view.getHistogramWidget()

        if self.data_type == DataType.normalization:
            self.parent.image_view_settings[data_type_of_data]["histogram"] = _histo_widget.getLevels()
        else:
            self.parent.image_view_settings[data_type_of_data]["histogram"][self.combine_algo] = (
                _histo_widget.getLevels()
            )

    def reload_histogram_level(self):
        if not self.first_update:
            if self.data_type in self.list_steps_without_combine_algo_options:
                if self.parent.image_view_settings[self.data_type]["histogram"]:
                    self.histo_widget.setLevels(
                        self.parent.image_view_settings[self.data_type]["histogram"][0],
                        self.parent.image_view_settings[self.data_type]["histogram"][1],
                    )
            else:
                if self.parent.image_view_settings[self.data_type]["histogram"][self.combine_algo]:
                    self.histo_widget.setLevels(
                        self.parent.image_view_settings[self.data_type]["histogram"][self.combine_algo][0],
                        self.parent.image_view_settings[self.data_type]["histogram"][self.combine_algo][1],
                    )

    def set_histogram_level(self, histogram_level):
        if histogram_level:
            if self.data_type in self.list_steps_without_combine_algo_options:
                self.histo_widget.setLevels(histogram_level[0], histogram_level[1])
                self.parent.image_view_settings[self.data_type]["histogram"] = histogram_level
            else:
                self.histo_widget.setLevels(histogram_level[0], histogram_level[1])
                self.parent.image_view_settings[self.data_type]["histogram"][self.combine_algo] = histogram_level



import numpy as np
from qtpy import QtCore

roi_group_color = [
    QtCore.Qt.darkBlue,
    QtCore.Qt.darkRed,
    QtCore.Qt.green,
    QtCore.Qt.darkYellow,
]


pen_color = {
    "0": (62, 13, 244),  # blue
    "1": (139, 10, 19),  # dark red
    "2": (36, 244, 31),  # green
    "3": (209, 230, 27),
}  # dark yellow


def set_alpha_value(lines=[], transparency=50):
    new_a = int((float(transparency) / 100.0) * 255)

    new_lines = []
    for _line in lines:
        [r, g, b, a, w] = _line
        _new_line = (r, g, b, new_a, w)
        new_lines.append(_new_line)

    lines = np.array(
        new_lines,
        dtype=[
            ("red", np.ubyte),
            ("green", np.ubyte),
            ("blue", np.ubyte),
            ("alpha", np.ubyte),
            ("width", float),
        ],
    )

    return lines



import pyqtgraph as pg
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QMainWindow,
    QTableWidgetItem,
    QTableWidgetSelectionRange,
)

from ibeatles import DEFAULT_ROI, DataType, load_ui
from ibeatles.step1.plot import Step1Plot
from ibeatles.utilities import colors
from ibeatles.utilities.gui_handler import GuiHandler


class RoiEditor:
    def __init__(self, parent=None):
        self.parent = parent

    def run(self):
        o_gui = GuiHandler(parent=self.parent)
        active_tab = o_gui.get_active_tab()

        list_roi_editor_ui = self.parent.roi_editor_ui
        _roi_ui = list_roi_editor_ui[active_tab]
        if _roi_ui is None:
            _interface = RoiEditorInterface(parent=self.parent, data_type=active_tab)
            _interface.show()

            # save ui id
            list_roi_editor_ui[active_tab] = _interface
            self.parent.roi_editor_ui = list_roi_editor_ui

        else:
            _interface = list_roi_editor_ui[active_tab]
            _interface.activateWindow()


class RoiEditorInterface(QMainWindow):
    col_width = [130, 35, 35, 43, 43]

    def __init__(self, parent=None, data_type=DataType.sample):
        self.parent = parent
        self.data_type = data_type

        QMainWindow.__init__(self, parent=parent)
        self.ui = load_ui("ui_roiEditor.ui", baseinstance=self)
        self.setWindowTitle("{} ROI Editor".format(data_type))

        self.initialize_table()
        self.fill_table()

    def initialize_table(self):
        for _index, _width in enumerate(self.col_width):
            self.ui.tableWidget.setColumnWidth(_index, _width)

        nbr_groups = len(colors.roi_group_color)
        self.list_name_groups = ["group {}".format(index) for index in range(nbr_groups)]

    def get_item(self, text, color):
        _item = QTableWidgetItem(str(text))
        _item.setForeground(color)

        return _item

    def fill_table(self):
        list_roi = self.parent.list_roi[self.data_type]

        # no ROI already define
        if len(list_roi) == 0:
            return

        self.ui.remove_roi_button.setEnabled(True)

        for _row, _roi in enumerate(list_roi):
            [label, x0, y0, width, height, group] = _roi
            self.ui.tableWidget.insertRow(_row)
            # _color = colors.roi_group_color[int(group)]
            self.set_row(_row, label, x0, y0, width, height, int(group))

        # QtCore.QObject.connect(self.ui.tableWidget, QtCore.SIGNAL("cellChanged(int, int)"), self.cell_changed)
        self.ui.tableWidget.cellChanged.connect(self.cell_changed)

    def set_row(self, _row, label, x0, y0, width, height, group):
        _color = colors.roi_group_color[int(group)]

        self.ui.tableWidget.blockSignals(True)

        # label
        _item = self.get_item(label, _color)
        self.ui.tableWidget.setItem(_row, 0, _item)

        # x0
        _item = self.get_item(x0, _color)
        self.ui.tableWidget.setItem(_row, 1, _item)

        # y0
        _item = self.get_item(y0, _color)
        self.ui.tableWidget.setItem(_row, 2, _item)

        # width
        _item = self.get_item(width, _color)
        self.ui.tableWidget.setItem(_row, 3, _item)

        # height
        _item = self.get_item(height, _color)
        self.ui.tableWidget.setItem(_row, 4, _item)

        # group
        _widget = QComboBox()
        _widget.addItems(self.list_name_groups)
        _widget.setCurrentIndex(int(group))
        _widget.currentIndexChanged.connect(self.changed_group)
        self.ui.tableWidget.setCellWidget(_row, 5, _widget)

        self.ui.tableWidget.blockSignals(False)

    def get_row(self, row=0):
        # label
        _item = self.ui.tableWidget.item(row, 0)
        if _item is None:
            raise ValueError
        label = str(_item.text())

        # x0
        _item = self.ui.tableWidget.item(row, 1)
        if _item is None:
            raise ValueError
        x0 = str(_item.text())

        # y0
        _item = self.ui.tableWidget.item(row, 2)
        if _item is None:
            raise ValueError
        y0 = str(_item.text())

        # width
        _item = self.ui.tableWidget.item(row, 3)
        if _item is None:
            raise ValueError
        width = str(_item.text())

        # height
        _item = self.ui.tableWidget.item(row, 4)
        if _item is None:
            raise ValueError
        height = str(_item.text())

        # group
        _group_widget = self.ui.tableWidget.cellWidget(row, 5)
        if _group_widget is None:
            raise ValueError
        _index_selected = _group_widget.currentIndex()
        group = str(_index_selected)

        return [label, x0, y0, width, height, group]

    def changed_group(self, _ignore):
        _nbr_row = self.ui.tableWidget.rowCount()
        for _row in range(_nbr_row):
            _group_widget = self.ui.tableWidget.cellWidget(_row, 5)
            _index_selected = _group_widget.currentIndex()

            _color = colors.roi_group_color[_index_selected]

            _item = self.ui.tableWidget.item(_row, 0)
            _item.setForeground(_color)

            _item = self.ui.tableWidget.item(_row, 1)
            _item.setForeground(_color)

            _item = self.ui.tableWidget.item(_row, 2)
            _item.setForeground(_color)

            _item = self.ui.tableWidget.item(_row, 3)
            _item.setForeground(_color)

            _item = self.ui.tableWidget.item(_row, 4)
            _item.setForeground(_color)

    def closeEvent(self, event=None):
        o_gui = GuiHandler(parent=self.parent)
        active_tab = o_gui.get_active_tab()
        self.parent.roi_editor_ui[active_tab] = None

    def add_roi_button_clicked(self):
        # self.ui.tableWidget.blockSignals(True)

        _row_selected = self.get_row_selected()
        if _row_selected == -1:
            _new_row_index = 0
        else:
            _new_row_index = _row_selected

        self.ui.tableWidget.insertRow(_new_row_index)

        list_roi = self.parent.list_roi[self.data_type]

        list_roi_id = self.parent.list_roi_id[self.data_type]
        list_label_roi_id = self.parent.list_label_roi_id[self.data_type]
        _nbr_row = len(list_roi)

        init_roi = DEFAULT_ROI
        [label, x0, y0, width, height, group] = init_roi
        x0 = int(x0)
        y0 = int(y0)
        width = int(width)
        height = int(height)

        # label roi
        label_roi = pg.TextItem(
            html='<div style="text-align: center"><span style="color: #ff0000;">' + label + "</span></div>",
            anchor=(-0.3, 1.3),
            border="w",
            fill=(0, 0, 255, 50),
        )

        # roi region in image
        roi = pg.ROI([x0, y0], [width, height])
        roi.addScaleHandle([1, 1], [0, 0])
        label_roi.setPos(x0, y0)
        if self.data_type == DataType.sample:
            self.parent.ui.image_view.addItem(roi)
            self.parent.ui.image_view.addItem(label_roi)
            roi.sigRegionChanged.connect(self.parent.roi_image_view_changed)
        elif self.data_type == DataType.ob:
            self.parent.ui.ob_image_view.addItem(roi)
            self.parent.ui.ob_image_view.addItem(label_roi)
            roi.sigRegionChanged.connect(self.parent.roi_ob_image_view_changed)
        elif self.data_type == DataType.normalized:
            self.parent.ui.normalized_image_view.addItem(roi)
            self.parent.ui.normalized_image_view.addItem(label_roi)
            roi.sigRegionChanged.connect(self.parent.roi_normalized_image_view_changed)

        list_roi.insert(0, init_roi)
        list_roi_id.insert(0, roi)
        list_label_roi_id.insert(0, label_roi)

        self.parent.list_roi[self.data_type] = list_roi
        self.parent.list_roi_id[self.data_type] = list_roi_id
        self.parent.list_label_roi_id[self.data_type] = list_label_roi_id

        # nbr_groups = len(colors.roi_group_color)
        # list_name_groups = ['group {}'.format(index) for index in range(nbr_groups)]

        # _color = colors.roi_group_color[0]
        _row = _new_row_index

        self.set_row(_row, label, x0, y0, width, height, int(group))
        self.ui.remove_roi_button.setEnabled(True)

        # self.ui.tableWidget.blockSignals(False)

    def remove_roi_button_clicked(self):
        _row_selected = self.get_row_selected()
        if _row_selected == -1:
            return

        self.ui.tableWidget.removeRow(_row_selected)

        list_roi = self.parent.list_roi[self.data_type]
        list_roi_id = self.parent.list_roi_id[self.data_type]
        list_label_roi_id = self.parent.list_label_roi_id[self.data_type]

        new_list_roi = []
        new_list_roi_id = []
        new_list_label_roi_id = []
        roi_to_remove = None
        for _index, _array in enumerate(list_roi):
            if _index == _row_selected:
                roi_to_remove = list_roi_id[_index]
                roi_label_to_remove = list_label_roi_id[_index]
                continue
            new_list_roi.append(_array)
            new_list_roi_id.append(list_roi_id[_index])
            new_list_label_roi_id.append(list_label_roi_id[_index])

        if len(new_list_roi) == 0:
            self.ui.remove_roi_button.setEnabled(False)

        self.parent.list_roi[self.data_type] = new_list_roi
        self.parent.list_roi_id[self.data_type] = new_list_roi_id
        self.parent.list_label_roi_id[self.data_type] = new_list_label_roi_id

        if self.data_type == DataType.sample:
            self.parent.ui.image_view.removeItem(roi_to_remove)
            self.parent.ui.image_view.removeItem(roi_label_to_remove)
        elif self.data_type == DataType.ob:
            self.parent.ui.ob_image_view.removeItem(roi_to_remove)
            self.parent.ui.ob_image_view.removeItem(roi_label_to_remove)
        elif self.data_type == DataType.normalized:
            self.parent.ui.normalized_image_view.removeItem(roi_to_remove)
            self.parent.ui.normalized_image_view.removeItem(roi_label_to_remove)

        self.cell_changed(0, 0)

    def get_row_selected(self):
        try:
            _row_selected = self.ui.tableWidget.selectedRanges()[0].bottomRow()
        except IndexError:
            _row_selected = -1

        return _row_selected

    def refresh(self, row):
        [label, x0, y0, width, height, group] = self.parent.list_roi[self.data_type][row]
        self.set_row(row, label, x0, y0, width, height, group)

    def activate_row(self, row):
        if row == -1:
            return

        nbr_column = self.ui.tableWidget.columnCount()
        nbr_row = self.ui.tableWidget.rowCount()
        full_range = QTableWidgetSelectionRange(0, 0, nbr_row, nbr_column - 1)
        self.ui.tableWidget.setRangeSelected(full_range, False)

        range_selected = QTableWidgetSelectionRange(row, 0, row, nbr_column - 1)
        self.ui.tableWidget.setRangeSelected(range_selected, True)
        QApplication.processEvents()

    def roi_editor_table_changed(self, row, column):
        _row = row
        try:
            row_variables = self.get_row(row=_row)
        except ValueError:
            return

        list_roi = self.parent.list_roi[self.data_type]
        list_roi[_row] = row_variables

        self.parent.list_roi[self.data_type] = list_roi

    def cell_changed(self, _i, _j):
        o_plot = Step1Plot(parent=self.parent, data_type=self.data_type)
        o_plot.display_bragg_edge(mouse_selection=False)

    def done_pushButton_clicked(self):
        self.close()



import os
import time

import numpy as np

from ibeatles import DataType
from ibeatles.step1.plot import Step1Plot


class RetrieveDataInfos:
    def __init__(self, parent=None, data_type="sample"):
        self.parent = parent
        self.data_type = data_type

        self.path = self.parent.data_metadata[data_type]["folder"]

        self.table_ui = {
            "sample": self.parent.ui.list_sample,
            "ob": self.parent.ui.list_open_beam,
            "normalized": self.parent.ui.list_normalized,
        }

        self.preview_widget = {
            "sample": self.parent.ui.image_view,
            "ob": self.parent.ui.image_view,
        }


class RetrieveGeneralDataInfos(RetrieveDataInfos):
    selected_infos = {
        "acquisition_duration": {"name": "Acquisition Duration", "value": 0},
        "acquisition_time": {"name": "Acquisition Time", "value": ""},
        "image_size": {"name": "Image(s) Size", "value": "512x512"},
        "image_type": {"name": "Image Type", "value": "16 bits"},
        "min_counts": {"name": "min counts", "value": 0},
        "max_counts": {"name": "max counts", "value": 0},
    }

    data = []

    def update(self, add_mean_radio_button_changed=False):
        list_row_selected = list(np.sort(self.get_list_row_selected()))

        if self.data_type == DataType.normalized:
            combine_algo = np.nansum if self.parent.ui.normalized_roi_add_button.isChecked() else np.nanmean
        else:
            combine_algo = np.nansum if self.parent.ui.roi_add_button.isChecked() else np.nanmean

        if not list_row_selected:
            if len(self.parent.data_metadata[self.data_type]["data"]) == 0:
                self.selected_infos = {}
                self.data = []
            else:
                list_files_selected = self.parent.list_file_selected[self.data_type]
                self.table_ui[self.data_type].blockSignals(True)
                for _row_selected in list_files_selected:
                    _item = self.table_ui[self.data_type].item(_row_selected)
                    _item.setSelected(True)
                self.table_ui[self.data_type].blockSignals(False)

                data = self.parent.data_metadata[self.data_type]["data"]

                self.data = combine_algo(data, axis=0)

        else:
            full_data = self.parent.data_metadata[self.data_type]["data"]
            _data = []
            for index in list_row_selected:
                _data.append(full_data[index])
            self.data = combine_algo(_data, axis=0)

            if self.data_type == "normalized":
                self.parent.data_metadata["normalized"]["data_live_selection"] = _data

            self.parent.list_file_selected[self.data_type] = list_row_selected

        self.display(add_mean_radio_button_changed=add_mean_radio_button_changed)

    def display(self, add_mean_radio_button_changed=False):
        # # metadata
        # text = ''
        # for key in self.selected_infos:
        #     text += '<b>{}</b>: {}<br/>'.format(self.selected_infos[key]['name'],
        #                                         self.selected_infos[key]['value'])
        # self.selected_infos_ui[self.data_type].setHtml(text)

        o_plot = Step1Plot(parent=self.parent, data_type=self.data_type, data=self.data)
        o_plot.display_image(add_mean_radio_button_changed=add_mean_radio_button_changed)

    def get_list_files_selected(self):
        list_files = [str(x.text()) for x in self.table_ui[self.data_type].selectedItems()]
        return list_files

    def get_list_row_selected(self):
        selection = self.table_ui[self.data_type].selectedIndexes()
        _list_row_selected = []
        for _index in selection:
            _list_row_selected.append(_index.row())
        return _list_row_selected


class RetrieveGeneralFileInfos(RetrieveDataInfos):
    general_infos = {
        "number_of_files": {"name": "Number of Files", "value": -1},
        "time_stamp_files": {"name": "Acquisition Time of First File", "value": -1},
        "size_mb": {"name": "Individual File Size (MB)", "value": ""},
        "total_size_folder": {"name": "Total Size of Folder (MB)", "value": ""},
        "full_path": {"name": "Full path to folder", "value": None},
        "folder": {"name": "Name of folder", "value": None},
    }

    def update(self):
        data_files = self.parent.list_files[self.data_type]

        if len(data_files) == 0:
            self.general_infos = {}  # no files so no infos to display

        else:
            folder = self.parent.data_metadata[self.data_type]["folder"]

            self.general_infos["full_path"]["value"] = folder
            self.general_infos["folder"]["value"] = os.path.basename(os.path.abspath(folder))

            _nbr_files = len(data_files)
            self.general_infos["number_of_files"]["value"] = _nbr_files

            _first_file = data_files[0]
            first_file_full_name = os.path.join(folder, _first_file)

            _timestamp_first_file = self.get_formated_time(first_file_full_name)
            self.general_infos["time_stamp_files"]["value"] = _timestamp_first_file

            _size_of_one_file_kb = float(os.path.getsize(first_file_full_name))
            _file_size_mb = "{:.2f}".format(_size_of_one_file_kb / 1000000.0)
            self.general_infos["size_mb"]["value"] = _file_size_mb

            _total_size_mb = _size_of_one_file_kb * _nbr_files / 1000000.0
            _total_size_mb = "{:.2f}".format(_total_size_mb)
            self.general_infos["total_size_folder"]["value"] = _total_size_mb

        # save general infos into main infos_dict
        self.parent.infos_dict[self.data_type] = self.general_infos

        # self.display()

    def get_formated_time(self, full_file_name):
        _time = time.strftime("%m/%d/%Y %H:%M:%S", time.gmtime(os.path.getmtime(full_file_name)))
        return _time

    # def display(self):
    #     text = ''
    #     for key in self.general_infos:
    #         text += '<b>{}</b>: {}<br/>'.format(self.general_infos[key]['name'],
    #                                             self.general_infos[key]['value'])
    #
    #     self.general_infos_ui[self.data_type].setHtml(text)



#!/usr/bin/env python
"""
Event handler
"""

import logging
import os

import numpy as np
import pandas as pd
import pyqtgraph as pg
from qtpy.QtWidgets import QFileDialog

from ibeatles import DataType, interact_me_style, normal_style

# MVP widget
from ibeatles.app.presenters.time_spectra_presenter import TimeSpectraPresenter

# backend function from core
from ibeatles.core.io.data_loading import get_shutter_count_filename, get_time_spectra_filename
from ibeatles.session import SessionSubKeys
from ibeatles.tools.tof_bin import BinAutoMode, BinMode
from ibeatles.tools.tof_bin.auto_event_handler import AutoEventHandler
from ibeatles.tools.tof_bin.manual_event_handler import ManualEventHandler
from ibeatles.tools.tof_bin.plot import Plot
from ibeatles.tools.tof_bin.utilities.get import Get
from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.utilities.file_handler import TIME_SPECTRA_NAME_FORMAT, FileHandler
from ibeatles.utilities.load_files import LoadFiles
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)
from ibeatles.utilities.table_handler import TableHandler


class EventHandler:
    def __init__(self, parent=None, top_parent=None):
        self.parent = parent
        self.top_parent = top_parent

        self.time_spectra_presenter = None

    def check_widgets(self):
        # enable widgets when a folder has been selected
        folder_selected = self.parent.ui.folder_selected.text()
        if os.path.exists(folder_selected):
            enabled_state = True
        else:
            enabled_state = False

        self.parent.ui.bin_tabWidget.setEnabled(enabled_state)
        self.parent.ui.x_axis_groupBox.setEnabled(enabled_state)
        self.parent.ui.stats_tabWidget.setEnabled(enabled_state)
        self.parent.ui.bin_bottom_tabWidget.setEnabled(enabled_state)
        self.parent.ui.export_pushButton.setEnabled(enabled_state)
        self.parent.ui.image_tabWidget.setEnabled(enabled_state)

        # enable export bin when there are bins selected
        o_get = Get(parent=self.parent)
        bin_mode = o_get.bin_mode()
        if bin_mode == BinMode.auto:
            state_auto_table_has_at_least_one_row_checked = self._auto_table_has_at_least_one_row_checked()
            self.parent.ui.export_pushButton.setEnabled(state_auto_table_has_at_least_one_row_checked)
        elif bin_mode == BinMode.manual:
            state_manual_table_has_at_least_one_real_bin = self._manual_table_has_at_least_one_real_bin()
            self.parent.ui.export_pushButton.setEnabled(state_manual_table_has_at_least_one_real_bin)

        # if data loaded, change stylesheet of buttons
        if self.parent.list_tif_files:
            self.parent.ui.select_folder_pushButton.setStyleSheet(normal_style)
            self.parent.ui.export_pushButton.setStyleSheet(interact_me_style)

    def _manual_table_has_at_least_one_real_bin(self):
        """return True if there is at least one bin defined"""

        if self.parent.manual_bins[TimeSpectraKeys.file_index_array]:
            if isinstance(self.parent.manual_bins[TimeSpectraKeys.file_index_array][0], list):
                return True
        return False

    def _auto_table_has_at_least_one_row_checked(self):
        """check that the auto table has at least one row enabled (first column widget is checked)"""
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        nbr_row = o_table.row_count()
        for _row in np.arange(nbr_row):
            widget = o_table.get_widget(row=_row, column=0)
            if widget:
                checkbox = widget.children()[1]
                if checkbox.isChecked():
                    return True
        return False

    def select_input_folder(self):
        default_path = self.top_parent.session_dict[DataType.sample][SessionSubKeys.current_folder]
        folder = QFileDialog.getExistingDirectory(
            parent=self.parent,
            caption="Select folder containing images to load",
            directory=default_path,
            options=QFileDialog.ShowDirsOnly,
        )

        if folder == "":
            logging.info("User Canceled the selection of folder!")
            show_status_message(
                parent=self.parent,
                message="User cancelled the file dialog window",
                duration_s=5,
                status=StatusMessageStatus.warning,
            )
            return

        list_tif_files = FileHandler.get_list_of_tif(folder=folder)
        if not list_tif_files:
            logging.info("-> folder does not contain any tif file!")
            show_status_message(
                parent=self.parent,
                message=f"Folder {os.path.basename(folder)} does not contain any TIFF files!",
                duration_s=5,
                status=StatusMessageStatus.error,
            )
            return

        self.parent.ui.folder_selected.setText(folder)
        logging.info(f"Users selected the folder: {folder}")
        self.parent.list_tif_files = list_tif_files

    def load_data(self):
        if not self.parent.list_tif_files:
            return

        dict = LoadFiles.load_interactive_data(parent=self.parent, list_tif_files=self.parent.list_tif_files)
        self.parent.image_size["height"] = dict["height"]
        self.parent.image_size["width"] = dict["width"]
        self.parent.images_array = dict["image_array"]
        self.parent.integrated_image = np.mean(dict["image_array"], axis=0)

    def load_shutter_counts_file(self, shutter_counts_file=None) -> bool:
        """
        load the shutter counts file content if there is one and return True
        otherwise return False
        """
        logging.info("Loading shutter counts file if there ...")
        if shutter_counts_file is None:
            folder = self.parent.ui.folder_selected.text()

            shutter_counts_file = get_shutter_count_filename(folder)
            logging.info(f"in load_shutter_counts_file, shutter_counts_file: {shutter_counts_file}")
            if not shutter_counts_file:
                logging.info("Shutter counts file not found!")
                return False

        # load file shutter count file
        full_shutter_data = pd.read_csv(shutter_counts_file, delim_whitespace=True, names=["ShutterCount"])
        # we keep only the counts > 0
        non_zero_shutter_counts = full_shutter_data[full_shutter_data["ShutterCount"] > 0].reset_index(drop=True)
        shutter_counts = non_zero_shutter_counts["ShutterCount"].tolist()
        self.parent.shutter_counts = shutter_counts
        self.parent.shutter_counts_file = shutter_counts_file
        self.parent.ui.shutter_counts_name_label.setText(os.path.basename(shutter_counts_file))

        return True

    def load_time_spectra_file(self):
        """
        load the time spectra file
        """
        logging.info("Loading time spectra file ...")
        folder = self.parent.ui.folder_selected.text()

        time_spectra_file = get_time_spectra_filename(folder)
        logging.info(f"in load_time_spectra_file, time_spectra_file: {time_spectra_file}")
        if not time_spectra_file:
            logging.info("Time spectra file not found!")
            show_status_message(
                parent=self.parent,
                message="Time spectra file not found!",
                status=StatusMessageStatus.error,
                duration_s=5,
            )
            return

        if self.time_spectra_presenter is None:
            self.time_spectra_presenter = TimeSpectraPresenter(self.parent)

        distance_source_detector_m = float(self.top_parent.ui.distance_source_detector.text())
        detector_offset = float(self.top_parent.ui.detector_offset.text())

        logging.info(f"distance_source_detector_m: {distance_source_detector_m}")
        logging.info(f"detector_offset: {detector_offset}")

        self.time_spectra_presenter.load_data(time_spectra_file, distance_source_detector_m, detector_offset)
        self.update_time_spectra_data()

        logging.info(f"time_spectra_file loaded: {time_spectra_file}")
        logging.info("time spectra data:")
        logging.info(f"\t {self.parent.time_spectra[TimeSpectraKeys.tof_array][0:5] =}")
        logging.info(f"\t {self.parent.time_spectra[TimeSpectraKeys.counts_array][0:5] =}")
        return True

    def browse_for_shutter_counts_file(self):
        """
        browse for shutter counts file then continue workflow
        (self.continue_import_workflow_after_time_spectra_loaded)
        """
        [shutter_counts_file, _] = QFileDialog.getOpenFileName(
            caption="Select the Shutter Counts File",
            directory=self.top_parent.default_path[DataType.sample],
            filter="Txt ({});;All (*.*)".format("*_ShutterCount.txt"),
        )
        if shutter_counts_file:
            is_shutter_counts_found = self.load_shutter_counts_file(shutter_counts_file)
            if is_shutter_counts_found:
                logging.info(f"shutter_counts_file loaded: {shutter_counts_file}")
                logging.info(f"shutter counts data: {self.parent.shutter_counts}")
            else:
                logging.error(f"shutter counts file not found: {shutter_counts_file}")
                show_status_message(
                    parent=self.parent,
                    message=f"Shutter count file not found: {shutter_counts_file}",
                    status=StatusMessageStatus.error,
                    duration_s=5,
                )
                self.top_parent.shutter_counts = None
                self.parent.ui.shutter_counts_name_label.setText("No file loaded")

    def browse_for_time_spectra_file(self):
        """browse for time spectra then continue workflow (self.continue_import_workflow_after_time_spectra_loaded)"""
        [time_spectra_file, _] = QFileDialog.getOpenFileName(
            caption="Select the Time Spectra File",
            directory=self.top_parent.default_path[DataType.sample],
            filter="Txt ({});;All (*.*)".format(TIME_SPECTRA_NAME_FORMAT),
        )
        if time_spectra_file:
            if self.time_spectra_presenter is None:
                self.time_spectra_presenter = TimeSpectraPresenter(self.parent)

                distance_source_detector_m = float(self.top_parent.ui.distance_source_detector.text())
                detector_offset = float(self.top_parent.ui.detector_offset.text())

                logging.info(f"distance_source_detector_m: {distance_source_detector_m}")
                logging.info(f"detector_offset: {detector_offset}")

                self.time_spectra_presenter.load_data(time_spectra_file, distance_source_detector_m, detector_offset)
                self.update_time_spectra_data()

                logging.info(f"time_spectra_file loaded: {time_spectra_file}")
                logging.info("time spectra data:")
                logging.info(f"\t {self.parent.time_spectra[TimeSpectraKeys.tof_array][0:5] =}")
                logging.info(f"\t {self.parent.time_spectra[TimeSpectraKeys.counts_array][0:5] =}")

    def update_time_spectra_data(self):
        logging.info("Updating time spectra data ...")
        time_spectra_data = self.time_spectra_presenter.model.get_data()

        self.parent.time_spectra[TimeSpectraKeys.file_name] = time_spectra_data["filename"]
        self.parent.time_spectra[TimeSpectraKeys.tof_array] = time_spectra_data["tof_array"]
        self.parent.time_spectra[TimeSpectraKeys.lambda_array] = time_spectra_data["lambda_array"]
        self.parent.time_spectra[TimeSpectraKeys.file_index_array] = np.arange(len(time_spectra_data["tof_array"]))
        self.parent.time_spectra[TimeSpectraKeys.counts_array] = time_spectra_data["counts_array"]

        # update time spectra tab
        self.parent.ui.time_spectra_name_label.setText(os.path.basename(time_spectra_data["filename"]))
        self.parent.ui.time_spectra_preview_pushButton.setEnabled(True)
        logging.info("... done!")

    def display_integrated_image(self):
        logging.info("Displaying integrated image ...")
        self.parent.integrated_view.clear()

        if self.parent.integrated_image is None:
            return

        self.parent.integrated_view.setImage(self.parent.integrated_image)

        roi = self.parent.bin_roi
        x0 = roi["x0"]
        y0 = roi["y0"]
        width = roi["width"]
        height = roi["height"]
        roi_item = pg.ROI([x0, y0], [width, height])
        roi_item.addScaleHandle([1, 1], [0, 0])
        self.parent.integrated_view.addItem(roi_item)
        roi_item.sigRegionChanged.connect(self.parent.bin_roi_changed)
        self.parent.roi_item = roi_item
        logging.info("... done!")

    def display_profile(self):
        logging.info("Displaying profile ...")
        if self.parent.integrated_image is None:
            return

        integrated_image = self.parent.integrated_image
        image_view = self.parent.integrated_view
        roi_item = self.parent.roi_item

        region = roi_item.getArraySlice(integrated_image, image_view.imageItem)
        x0 = region[0][0].start
        x1 = region[0][0].stop - 1
        y0 = region[0][1].start
        y1 = region[0][1].stop - 1

        width = x1 - x0
        height = y1 - y0

        self.parent.bin_roi = {"x0": x0, "y0": y0, "width": width, "height": height}

        o_plot = Plot(parent=self.parent)
        o_plot.refresh_profile_plot_and_clear_bins()
        logging.info("... done!")

    def bin_xaxis_changed(self):
        o_get = Get(parent=self.parent)
        if o_get.bin_mode() == BinMode.auto:
            self.entering_tab_auto()
        elif o_get.bin_mode() == BinMode.manual:
            o_event = ManualEventHandler(parent=self.parent)
            # o_event.update_manual_snapping_indexes_bins()
            o_event.update_items_displayed()

    def bin_auto_manual_tab_changed(self, new_tab_index=-1):
        if new_tab_index == -1:
            new_tab_index = self.parent.ui.bin_tabWidget.currentIndex()

        if new_tab_index == 0:
            self.parent.session[SessionSubKeys.bin_mode] = BinMode.auto

        elif new_tab_index == 1:
            self.parent.session[SessionSubKeys.bin_mode] = BinMode.manual

        elif new_tab_index == 2:
            pass

        else:
            raise NotImplementedError("LinearBin mode not implemented!")

        self.entering_tab()

    def entering_tab(self):
        o_get = Get(parent=self.parent)
        if o_get.bin_mode() == BinMode.auto:
            self.entering_tab_auto()
        elif o_get.bin_mode() == BinMode.manual:
            self.entering_tab_manual()
        else:
            raise NotImplementedError("tab not implemented yet!")

    def entering_tab_manual(self):
        # if table has one entry, select first row
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        nbr_row = o_table.row_count()
        if nbr_row > 0:
            o_table.select_row(0)

        o_plot = Plot(parent=self.parent)
        o_plot.refresh_profile_plot_and_clear_bins()

        o_manual_event = ManualEventHandler(parent=self.parent)
        o_manual_event.display_all_items()

    def entering_tab_auto(self):
        o_get = Get(parent=self.parent)
        o_auto_event = AutoEventHandler(parent=self.parent)
        if o_get.bin_auto_mode() == BinAutoMode.linear:
            o_auto_event.auto_linear_radioButton_changed()
        elif o_get.bin_auto_mode() == BinAutoMode.log:
            o_auto_event.auto_log_radioButton_changed()
        o_auto_event.refresh_auto_tab()



#!/usr/bin/env python
"""
Get module
"""

import copy
import os
from os.path import expanduser
from pathlib import Path

import numpy as np
import tomli

from ibeatles import DataType
from ibeatles.tools.tof_combine import SessionKeys as TofSessionKeys
from ibeatles.tools.tof_combine.utilities.table_handler import TableHandler
from ibeatles.tools.utilities import CombineAlgorithm, TimeSpectraKeys


class Get:
    def __init__(self, parent=None):
        self.parent = parent

    def combine_algorithm(self):
        if self.parent.ui.combine_mean_radioButton.isChecked():
            return CombineAlgorithm.mean
        elif self.parent.ui.combine_median_radioButton.isChecked():
            return CombineAlgorithm.median
        else:
            raise NotImplementedError("Combine algorithm not implemented!")

    def combine_x_axis_selected(self):
        if self.parent.combine_file_index_radio_button.isChecked():
            return TimeSpectraKeys.file_index_array
        elif self.parent.tof_radio_button.isChecked():
            return TimeSpectraKeys.tof_array
        elif self.parent.lambda_radio_button.isChecked():
            return TimeSpectraKeys.lambda_array
        else:
            raise NotImplementedError("xaxis not implemented in the combine tab!")

    def list_of_folders_status(self):
        """return a dict of the folder status (True if the checkbox is checked)"""
        o_table = TableHandler(table_ui=self.parent.ui.combine_tableWidget)
        nbr_row = o_table.row_count()
        dict_folders_status = {}
        for _row_index in np.arange(nbr_row):
            _horizontal_widget = o_table.get_widget(row=_row_index, column=0)
            radio_button = _horizontal_widget.layout().itemAt(1).widget()
            if radio_button.isChecked():
                dict_folders_status[_row_index] = True
            else:
                dict_folders_status[_row_index] = False

        return dict_folders_status

    def number_of_folders_we_want_to_combine(self):
        """return the number of folders with the radio button checked (we want to be part of the combine)"""
        nbr_folders = 0
        o_table = TableHandler(table_ui=self.parent.ui.combine_tableWidget)
        nbr_row = o_table.row_count()
        for _row_index in np.arange(nbr_row):
            _horizontal_widget = o_table.get_widget(row=_row_index, column=0)
            radio_button = _horizontal_widget.layout().itemAt(1).widget()
            if radio_button.isChecked():
                nbr_folders += 1

        return nbr_folders

    def row_of_that_folder(self, folder):
        """returns the row number where the folder has been found"""
        for _row in self.parent.dict_data_folders.keys():
            if self.parent.dict_data_folders[_row][TofSessionKeys.folder] == folder:
                return _row
        return -1

    def list_array_to_combine(self):
        """return the list of array to combine (list folders with checkbox checked"""

        list_row_to_use = []
        for _row in self.parent.dict_data_folders.keys():
            if self.parent.dict_data_folders[_row][TofSessionKeys.use]:
                list_row_to_use.append(_row)

        if not list_row_to_use:
            return None

        list_array = []
        for _row in list_row_to_use:
            list_array.append(copy.deepcopy(self.parent.dict_data_folders[_row][TofSessionKeys.data]))

        return list_array

    def manual_working_row(self, working_item_id=None):
        list_item_id = self.parent.list_of_manual_bins_item
        for _row, item in enumerate(list_item_id):
            if item == working_item_id:
                return _row
        return -1

    def combine_export_mode(self):
        if self.parent.ui.none_radioButton.isChecked():
            return DataType.none
        elif self.parent.ui.sample_radioButton.isChecked():
            return DataType.sample
        elif self.parent.ui.ob_radioButton.isChecked():
            return DataType.ob
        elif self.parent.ui.normalized_radioButton.isChecked():
            return DataType.normalized
        else:
            raise NotImplementedError("export mode not implemented yet!")

    @staticmethod
    def full_home_file_name(base_file_name):
        home_folder = expanduser("~")
        full_log_file_name = os.path.join(home_folder, base_file_name)
        return full_log_file_name

    @staticmethod
    def version():
        setup_cfg = "pyproject.toml"
        this_folder = os.path.abspath(os.path.dirname(__file__))
        top_path = Path(this_folder).parent.parent
        full_path_setup_cfg = str(Path(top_path) / Path(setup_cfg))

        ## to read from pyproject.toml file
        with open(full_path_setup_cfg, "rb") as fp:
            config = tomli.load(fp)
        version = config["project"]["version"]

        ## to read from setup.cfg file
        # config = configparser.ConfigParser()
        # config.read(full_path_setup_cfg)
        # version = config['project']['version']
        return version



from ibeatles.core.io.data_loading import load_time_spectra


class TimeSpectraModel:
    def __init__(self):
        self.data = {}

    def load_data(
        self,
        file_path: str,
        distance_source_detector_m: float,
        detector_offset_micros: float,
    ):
        self.data = load_time_spectra(file_path, distance_source_detector_m, detector_offset_micros)

    def get_data(self):
        return self.data



import platform


def get_platform():
    return platform.platform()


def is_os_mac():
    if "macos" in get_platform().lower():
        return True
    return False


def is_os_linux():
    if "linux" in get_platform().lower():
        return True
    return False



#!/usr/bin/env python
"""
Load fitting tab
"""

from ibeatles import DataType
from ibeatles.session import SessionKeys, SessionSubKeys


class LoadFitting:
    def __init__(self, parent=None):
        self.parent = parent
        self.session_dict = parent.session_dict

    def table_dictionary(self):
        self.parent.session_dict[SessionKeys.fitting] = self.session_dict[SessionKeys.fitting]
        self.parent.table_loaded_from_session = True

        self.parent.image_view_settings[DataType.fitting]["state"] = self.parent.session_dict[DataType.fitting][
            SessionSubKeys.image_view_state
        ]
        self.parent.image_view_settings[DataType.fitting]["histogram"] = self.parent.session_dict[DataType.fitting][
            SessionSubKeys.image_view_histogram
        ]



#!/usr/bin/env python
"""Export functions for strain mapping analysis results."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional

import numpy as np
import pandas as pd
from lmfit.model import ModelResult
from matplotlib.figure import Figure

from ibeatles.core.config import (
    BinCoordinates,
    OutputFileConfig,
)

logger = logging.getLogger(__name__)


def generate_output_filename(
    input_folder: str | Path,
    analysis_type: str,
    extension: str,
    timestamp: Optional[datetime] = None,
) -> Path:
    """Generate standardized output filename with timestamp.

    Parameters
    ----------
    input_folder : str | Path
        Original input data folder name to use as base
    analysis_type : str
        Type of analysis result (e.g., 'strain_map', 'fitting_grid')
    extension : str
        File extension without dot (e.g., 'png', 'pdf', 'csv')
    timestamp : Optional[datetime]
        Timestamp to use, defaults to current time if None

    Returns
    -------
    Path
        Complete output filepath

    Example
    -------
    >>> generate_output_filename('fe_sample', 'strain_map', 'png')
    Path('fe_sample_strain_map_20240108_153042.png')
    """
    input_folder = Path(input_folder)
    base_name = input_folder.stem

    if timestamp is None:
        timestamp = datetime.now()

    timestamp_str = timestamp.strftime("%Y%m%d_%H%M%S")
    filename = f"{base_name}_{analysis_type}_{timestamp_str}.{extension}"

    return Path(filename)


def save_strain_map(
    figure: Figure,
    output_path: Path,
    config: OutputFileConfig,
) -> None:
    """Save strain map overlay plot.

    Parameters
    ----------
    figure : Figure
        Matplotlib figure containing the strain map
    output_path : Path
        Path to save the figure
    config : OutputFileConfig
        Output configuration including format and DPI settings

    Raises
    ------
    ValueError
        If output path already exists
    IOError
        If saving fails
    """
    output_path = Path(output_path)
    if output_path.exists():
        raise ValueError(f"Output path {output_path} already exists")

    try:
        figure.savefig(
            output_path,
            format=config.strain_map_format,
            dpi=config.figure_dpi,
            bbox_inches="tight",
        )
        logger.info(f"Saved strain map to {output_path}")
    except Exception as e:
        logger.error(f"Failed to save strain map: {str(e)}")
        raise IOError(f"Failed to save strain map: {str(e)}")


def save_fitting_grid(
    figure: Figure,
    output_path: Path,
    config: OutputFileConfig,
) -> None:
    """Save fitting results grid plot.

    Parameters
    ----------
    figure : Figure
        Matplotlib figure containing the fitting grid
    output_path : Path
        Path to save the figure
    config : OutputFileConfig
        Output configuration including format and DPI settings

    Raises
    ------
    ValueError
        If output path already exists
    IOError
        If saving fails
    """
    output_path = Path(output_path)
    if output_path.exists():
        raise ValueError(f"Output path {output_path} already exists")

    try:
        figure.savefig(
            output_path,
            format=config.fitting_grid_format,
            dpi=config.figure_dpi,
            bbox_inches="tight",
        )
        logger.info(f"Saved fitting grid to {output_path}")
    except Exception as e:
        logger.error(f"Failed to save fitting grid: {str(e)}")
        raise IOError(f"Failed to save fitting grid: {str(e)}")


def save_analysis_results(
    fit_results: Dict[str, Optional[ModelResult]],
    bin_coordinates: Dict[str, BinCoordinates],
    strain_results: Dict[str, Dict[str, float]],
    metadata: Dict[str, Any],
    output_path: Path,
    config: OutputFileConfig,
) -> pd.DataFrame:
    """Save fitting and strain results as CSV with metadata header.

    Parameters
    ----------
    fit_results : Dict[str, Optional[ModelResult]]
        Dictionary of fitting results for each bin
    bin_coordinates : Dict[str, BinCoordinates]
        Dictionary of bin coordinates
    strain_results : Dict[str, Dict[str, float]]
        Dictionary of strain calculation results
    metadata : Dict[str, Any]
        Analysis metadata to include in header
    output_path : Path
        Path to save the CSV file
    config : OutputFileConfig
        Output configuration including CSV format settings

    Returns
    -------
    pd.DataFrame
        DataFrame containing the analysis results

    Raises
    ------
    ValueError
        If output path already exists
    IOError
        If saving fails
    """
    output_path = Path(output_path)
    if output_path.exists():
        raise ValueError(f"Output path {output_path} already exists")

    # Prepare metadata header
    header_lines = []
    if config.csv_format.include_metadata_header:
        comment_char = config.csv_format.metadata_comment_char
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        header_lines.extend(
            [
                f"{comment_char} Analysis Timestamp: {timestamp}",
                f"{comment_char} Material: {metadata.get('material_name', 'Unknown')}",
                f"{comment_char} d0: {metadata.get('d0', 'Unknown')} Å",
                f"{comment_char} Distance Source-Detector: {metadata.get('distance_source_detector', 'Unknown')} m",
                f"{comment_char} Detector Offset: {metadata.get('detector_offset', 'Unknown')} µs",
                f"{comment_char}",
            ]
        )

    # Create DataFrame
    data = []
    for bin_id in fit_results.keys():
        result = fit_results[bin_id]
        coords = bin_coordinates[int(bin_id)]
        strain_data = strain_results.get(bin_id, {})

        if result is not None:
            # extract fitting results
            # NOTE: sometimes the uncertainty estimate fails due to a, b, sigma, tau at the
            #       boundaries of the parameter space. In this case, the stderr is NaN.
            lambda_hkl = result.best_values.get("bragg_edge_wavelength", np.nan)
            lambda_hkl_err = result.params["bragg_edge_wavelength"].stderr
            d_spacing = lambda_hkl / 2.0
            d_spacing_err = lambda_hkl_err / 2.0 if lambda_hkl_err else np.nan
            row = {
                "bin_id": bin_id,
                "x0": coords.x0,
                "y0": coords.y0,
                "x1": coords.x1,
                "y1": coords.y1,
                "row_index": coords.row_index,
                "column_index": coords.column_index,
                "lambda_hkl": lambda_hkl,
                "lambda_hkl_err": lambda_hkl_err,
                "d_spacing": d_spacing,
                "d_spacing_err": d_spacing_err,
                "strain": strain_data.get("strain", np.nan),
                "strain_err": strain_data.get("error", np.nan),
                "r_squared": result.rsquared,
                "chi_squared": result.chisqr,
            }
            data.append(row)

    df = pd.DataFrame(data)

    try:
        # Write header lines first if enabled
        if config.csv_format.include_metadata_header and header_lines:
            with open(output_path, "w") as f:
                f.write("\n".join(header_lines) + "\n")

            # Append data
            df.to_csv(
                output_path,
                mode="a",
                index=False,
                sep=config.csv_format.delimiter,
                na_rep=config.csv_format.na_rep,
            )
        else:
            # Write data directly
            df.to_csv(
                output_path,
                index=False,
                sep=config.csv_format.delimiter,
            )

        logger.info(f"Saved analysis results to {output_path}")
    except Exception as e:
        logger.error(f"Failed to save analysis results: {str(e)}")
        raise IOError(f"Failed to save analysis results: {str(e)}")

    return df



#!/usr/bin/env python
"""
Event handler module
"""

import copy
import json
import logging
from pathlib import PurePath

import numpy as np
from qtpy import QtGui
from qtpy.QtWidgets import QApplication, QMenu

from ibeatles import DataType, interact_me_style, normal_style
from ibeatles.fitting import FittingKeys, FittingTabSelected, KropffTabSelected
from ibeatles.fitting.fitting_handler import FittingHandler
from ibeatles.fitting.get import Get
from ibeatles.fitting.kropff import (
    UNLOCK_ROW_BACKGROUND,
    BraggPeakInitParameters,
    FittingKropffBraggPeakColumns,
    FittingKropffHighLambdaColumns,
    FittingKropffLowLambdaColumns,
)
from ibeatles.fitting.kropff import SessionSubKeys as KropffSessionSubKeys
from ibeatles.fitting.kropff.checking_fitting_conditions import (
    CheckingFittingConditions,
)
from ibeatles.fitting.kropff.display import Display
from ibeatles.fitting.kropff.fit_regions import FitRegions
from ibeatles.fitting.kropff.fitting_parameters_viewer_editor_launcher import (
    FittingParametersViewerEditorLauncher,
)
from ibeatles.fitting.kropff.get import Get as KropffGet
from ibeatles.fitting.kropff.kropff_bragg_peak_threshold_calculator import (
    KropffBraggPeakThresholdCalculator,
)
from ibeatles.session import SessionSubKeys
from ibeatles.utilities.array_utilities import from_nparray_to_list
from ibeatles.utilities.check import is_float, is_nan
from ibeatles.utilities.file_handler import select_folder
from ibeatles.utilities.status_message_config import (
    StatusMessageStatus,
    show_status_message,
)
from ibeatles.utilities.table_handler import TableHandler

fit_rgb = (255, 0, 0)


class EventHandler:
    default_threshold_width = 3

    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def reset_fitting_parameters(self):
        table_dictionary = self.grand_parent.kropff_table_dictionary

        kropff_table_dictionary_template = FittingHandler.kropff_table_dictionary_template
        for _row in table_dictionary.keys():
            for _template_key in kropff_table_dictionary_template.keys():
                table_dictionary[_row][_template_key] = copy.deepcopy(kropff_table_dictionary_template[_template_key])

    def _is_first_row_has_threshold_defined(self):
        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        kropff_table_of_second_row = kropff_table_dictionary["1"]

        if kropff_table_of_second_row["bragg peak threshold"]["left"] is None:
            return False

        if kropff_table_of_second_row["yaxis"] is None:
            self.record_all_xaxis_and_yaxis()
            if kropff_table_of_second_row["yaxis"] is None:
                return False

        return True

    def _we_are_ready_to_fit_all_regions(self):
        return self._is_first_row_has_threshold_defined()

    def check_widgets_helper(self):
        if self._we_are_ready_to_fit_all_regions():
            self.parent.ui.kropff_fit_allregions_pushButton.setEnabled(True)
            self.parent.ui.kropff_fit_allregions_pushButton.setStyleSheet(interact_me_style)
            self.parent.ui.automatic_bragg_peak_threshold_finder_pushButton.setStyleSheet(normal_style)
        else:
            self.parent.ui.kropff_fit_allregions_pushButton.setEnabled(False)
            self.parent.ui.kropff_fit_allregions_pushButton.setStyleSheet(normal_style)
            self.parent.ui.automatic_bragg_peak_threshold_finder_pushButton.setStyleSheet(interact_me_style)

    def record_all_xaxis_and_yaxis(self):
        table_dictionary = self.grand_parent.kropff_table_dictionary
        nbr_row = len(table_dictionary.keys())
        data_2d = self.grand_parent.data_metadata["normalized"]["data"]

        # index of selection in bragg edge plot
        [left_index, right_index] = self.grand_parent.fitting_bragg_edge_linear_selection
        full_x_axis = self.parent.bragg_edge_data["x_axis"]
        xaxis = np.array(full_x_axis[left_index:right_index], dtype=float)

        for _row in np.arange(nbr_row):
            _bin_entry = table_dictionary[str(_row)]

            _bin_x0 = _bin_entry["bin_coordinates"]["x0"]
            _bin_x1 = _bin_entry["bin_coordinates"]["x1"]
            _bin_y0 = _bin_entry["bin_coordinates"]["y0"]
            _bin_y1 = _bin_entry["bin_coordinates"]["y1"]

            yaxis = data_2d[
                left_index:right_index,
                _bin_x0:_bin_x1,
                _bin_y0:_bin_y1,
            ]  # noqa: E124
            yaxis = np.nanmean(yaxis, axis=1)
            yaxis = np.array(np.nanmean(yaxis, axis=1), dtype=float)
            _bin_entry["yaxis"] = yaxis
            _bin_entry["xaxis"] = xaxis

    def update_fitting_plots(self):
        self.update_bottom_right_plot()

    def update_bottom_right_plot(self):
        self.parent.ui.kropff_fitting.clear()

        o_get = Get(parent=self.parent, grand_parent=self.grand_parent)
        yaxis, xaxis = o_get.y_axis_and_x_axis_for_given_rows_selected()

        self.parent.ui.kropff_fitting.setLabel("left", "Cross Section (arbitrary units, -log(counts))")
        self.parent.ui.kropff_fitting.setLabel("bottom", "\u03bb (\u212b)")

        for _yaxis in yaxis:
            _yaxis = -np.log(_yaxis)
            self.parent.ui.kropff_fitting.plot(xaxis, _yaxis, symbol="o", pen=None)

        xaxis_fitted, yaxis_fitted = o_get.y_axis_fitted_for_given_rows_selected()

        if yaxis_fitted:
            for _yaxis in yaxis_fitted:
                self.parent.ui.kropff_fitting.plot(xaxis_fitted, _yaxis, pen=(fit_rgb[0], fit_rgb[1], fit_rgb[2]))

        o_display = Display(parent=self.parent, grand_parent=self.grand_parent)
        o_display.display_bragg_peak_threshold()

    def kropff_automatic_bragg_peak_threshold_finder_clicked(self):
        o_kropff = KropffBraggPeakThresholdCalculator(parent=self.parent, grand_parent=self.grand_parent)
        o_kropff.save_all_profiles()
        o_kropff.run_automatic_mode()

        o_display = Display(parent=self.parent, grand_parent=self.grand_parent)
        o_display.display_bragg_peak_threshold()
        self.parent.ui.kropff_fit_allregions_pushButton.setEnabled(True)

    def kropff_bragg_edge_threshold_changed(self):
        lr = self.parent.kropff_threshold_current_item
        [left, right] = lr.getRegion()

        o_kropff = Get(parent=self.parent)
        row_selected = str(o_kropff.kropff_row_selected()[0])

        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        kropff_table_of_row_selected = kropff_table_dictionary[row_selected]
        kropff_table_of_row_selected["bragg peak threshold"]["left"] = left
        kropff_table_of_row_selected["bragg peak threshold"]["right"] = right

    def parameters_changed(self):
        # high TOF
        a0 = self.parent.ui.kropff_high_lda_a0_init.text()
        b0 = self.parent.ui.kropff_high_lda_b0_init.text()
        high_tof_graph = "a0" if self.parent.ui.kropff_a0_radioButton.isChecked() else "b0"
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.high_tof][
            KropffSessionSubKeys.a0
        ] = a0
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.high_tof][
            KropffSessionSubKeys.b0
        ] = b0
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.high_tof][
            KropffSessionSubKeys.graph
        ] = high_tof_graph

        # low TOF
        ahkl = self.parent.ui.kropff_low_lda_ahkl_init.text()
        bhkl = self.parent.ui.kropff_low_lda_bhkl_init.text()
        low_tof_graph = "ahkl" if self.parent.ui.kropff_ahkl_radioButton.isChecked() else "bhkl"
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.low_tof][
            KropffSessionSubKeys.ahkl
        ] = ahkl
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.low_tof][
            KropffSessionSubKeys.bhkl
        ] = bhkl
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.low_tof][
            KropffSessionSubKeys.graph
        ] = low_tof_graph

        # bragg peak
        lambda_hkl_fix_flag = self.parent.ui.lambda_hkl_fix_radioButton.isChecked()
        lambda_hkl_fix_value = self.parent.ui.lambda_hkl_fix_lineEdit.text()
        lambda_hkl_range_from = self.parent.ui.lambda_hkl_from_lineEdit.text()
        lambda_hkl_range_to = self.parent.ui.lambda_hkl_to_lineEdit.text()
        lambda_hkl_range_step = self.parent.ui.lambda_hkl_step_lineEdit.text()

        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.lambda_hkl
        ][BraggPeakInitParameters.fix_value] = lambda_hkl_fix_value
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.lambda_hkl
        ][BraggPeakInitParameters.fix_flag] = lambda_hkl_fix_flag
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.lambda_hkl
        ][BraggPeakInitParameters.range_from] = lambda_hkl_range_from
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.lambda_hkl
        ][BraggPeakInitParameters.range_to] = lambda_hkl_range_to
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.lambda_hkl
        ][BraggPeakInitParameters.range_step] = lambda_hkl_range_step

        # parameters required by strain mapping window
        from_lambda = self.parent.ui.lambda_min_lineEdit.text()
        to_lambda = self.parent.ui.lambda_max_lineEdit.text()
        hkl_selected = self.parent.ui.hkl_list_ui.currentText()
        lambda_0 = float(self.parent.ui.bragg_edge_calculated.text())
        element = self.parent.ui.material_groupBox.title()

        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.from_lambda
        ] = from_lambda
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.to_lambda
        ] = to_lambda
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.hkl_selected
        ] = hkl_selected
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            BraggPeakInitParameters.lambda_0
        ] = lambda_0
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][BraggPeakInitParameters.element] = (
            element
        )

        # bragg peak parameters
        tau_fix_flag = self.parent.ui.tau_fix_radioButton.isChecked()
        tau_fix_value = self.parent.ui.tau_fix_lineEdit.text()
        tau_range_from = self.parent.ui.tau_from_lineEdit.text()
        tau_range_to = self.parent.ui.tau_to_lineEdit.text()
        tau_range_step = self.parent.ui.tau_step_lineEdit.text()
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.tau
        ][BraggPeakInitParameters.fix_value] = tau_fix_value
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.tau
        ][BraggPeakInitParameters.fix_flag] = tau_fix_flag
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.tau
        ][BraggPeakInitParameters.range_from] = tau_range_from
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.tau
        ][BraggPeakInitParameters.range_to] = tau_range_to
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.tau
        ][BraggPeakInitParameters.range_step] = tau_range_step

        sigma_fix_flag = self.parent.ui.sigma_fix_radioButton.isChecked()
        sigma_fix_value = self.parent.ui.sigma_fix_lineEdit.text()
        sigma_range_from = self.parent.ui.sigma_from_lineEdit.text()
        sigma_range_to = self.parent.ui.sigma_to_lineEdit.text()
        sigma_range_step = self.parent.ui.sigma_step_lineEdit.text()
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.sigma
        ][BraggPeakInitParameters.fix_value] = sigma_fix_value
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.sigma
        ][BraggPeakInitParameters.fix_flag] = sigma_fix_flag
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.sigma
        ][BraggPeakInitParameters.range_from] = sigma_range_from
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.sigma
        ][BraggPeakInitParameters.range_to] = sigma_range_to
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.sigma
        ][BraggPeakInitParameters.range_step] = sigma_range_step

        # lambda_hkl = self.parent.kropff_lambda_settings['fix']
        # tau = self.parent.ui.kropff_bragg_peak_tau_init.text()
        # sigma = self.parent.ui.kropff_bragg_peak_sigma_comboBox.currentText()
        if self.parent.ui.kropff_lda_hkl_radioButton.isChecked():
            bragg_peak_graph = "lambda_hkl"
        elif self.parent.ui.kropff_tau_radioButton.isChecked():
            bragg_peak_graph = "tau"
        else:
            bragg_peak_graph = "sigma"

        # self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
        #     KropffSessionSubKeys.lambda_hkl] = lambda_hkl
        # self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
        #     KropffSessionSubKeys.tau] = tau
        # self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
        #     KropffSessionSubKeys.sigma] = sigma
        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][KropffSessionSubKeys.bragg_peak][
            KropffSessionSubKeys.graph
        ] = bragg_peak_graph

        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            KropffSessionSubKeys.kropff_bragg_peak_good_fit_conditions
        ] = self.parent.kropff_bragg_peak_good_fit_conditions

        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            KropffSessionSubKeys.kropff_lambda_settings
        ] = self.parent.kropff_lambda_settings

        self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            KropffSessionSubKeys.bragg_peak_row_rejections_conditions
        ] = self.parent.kropff_bragg_peak_row_rejections_conditions

    def fit_regions(self):
        o_fit = FitRegions(parent=self.parent, grand_parent=self.grand_parent)
        o_fit.all_regions()

    def fit_bragg_peak(self):
        o_fit = FitRegions(parent=self.parent, grand_parent=self.grand_parent)
        o_fit.bragg_peak()

    def bragg_peak_right_click(self):
        menu = QMenu(self.parent)

        # lock_all_good_cells = None
        # unlock_all_rows = None
        # replace_row = None
        # export_bin = None

        export_bin = menu.addAction("Export bin signal ...")

        # replace_row = menu.addAction("Replace value by median of surrounding pixels")
        # button_state = self.parent.kropff_bragg_table_right_click_menu[RightClickTableMenu.replace_values]['state']
        # replace_row.setEnabled(button_state)

        menu.addSeparator()

        display_fitting_parameters = menu.addAction("Fitting parameters viewer ...")
        button_state = self.did_we_perform_the_fitting()

        display_fitting_parameters.setEnabled(button_state)
        # unlock_all_rows = menu.addAction("Un-lock/Un-reject all rows")

        # self.parent.kropff_bragg_table_right_click_menu[RightClickTableMenu.replace_values]['ui'] = replace_row
        # self.parent.kropff_bragg_table_right_click_menu[RightClickTableMenu.display_fitting_parameters]['ui'] = \
        #     display_fitting_parameters

        action = menu.exec_(QtGui.QCursor.pos())

        # if action == unlock_all_rows:
        #     self.unlock_all_bragg_peak_rows()
        # if action == replace_row:
        #     self.replace_bragg_peak_row_values()
        if action == display_fitting_parameters:
            self.display_fitting_parameters()
        elif action == export_bin:
            self.export_bin()

    def did_we_perform_the_fitting(self) -> bool:
        """
        do we have the fitting table filled with values
         if at least one lambda_hkl value is found, return True, otherwise return False

        Returns
            bool
        """
        table = self.grand_parent.kropff_table_dictionary
        for _row in table.keys():
            lambda_hkl_value = table[_row][KropffSessionSubKeys.lambda_hkl]["val"]
            if is_nan(lambda_hkl_value):
                continue

            if is_float(lambda_hkl_value):
                return True
        return False

    def export_bin(self):
        logging.info("Exporting bin:")
        o_table = TableHandler(table_ui=self.parent.ui.bragg_edge_tableWidget)
        row_selected = str(o_table.get_row_selected())

        # create base output file name using bin# row# col#
        table_dictionary = self.grand_parent.kropff_table_dictionary
        metadata_for_this_row = table_dictionary[row_selected]

        logging.info(f" - row_selected: {row_selected}")
        # logging.info(f" - metadata: {metadata_for_this_row}")

        bin_number = row_selected
        row_index = metadata_for_this_row[FittingKeys.row_index]
        column_index = metadata_for_this_row[FittingKeys.column_index]

        logging.info(f" - bin_number: {bin_number}")
        logging.info(f" - bin row: {row_index}")
        logging.info(f" - bin column: {column_index}")

        # {'high_tof': {'xaxis': None, 'yaxis': None},
        #  'low_tof': {'xaxis': None, 'yaxis': None},
        #  'bragg_peak': {'xaxis': None, 'yaxis': None},
        fitted_dict = metadata_for_this_row[KropffSessionSubKeys.fitted]

        parent_folder = PurePath(self.grand_parent.default_path[DataType.normalized]).parent
        base_parent_folder = PurePath(self.grand_parent.default_path[DataType.normalized]).name
        output_folder = select_folder(start_folder=str(parent_folder))

        full_output_filename = PurePath(output_folder) / f"{str(base_parent_folder)}_bin#{bin_number}.json"
        logging.info(f" - output file name: {full_output_filename}")

        # cleanup data
        cleaned_dict = {}

        cleaned_dict[FittingKeys.x_axis] = from_nparray_to_list(metadata_for_this_row[FittingKeys.x_axis])
        cleaned_dict[FittingKeys.y_axis] = from_nparray_to_list(metadata_for_this_row[FittingKeys.y_axis])

        # fitted
        cleaned_dict[KropffSessionSubKeys.fitted] = {}

        # high tof
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.high_tof] = {}
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.high_tof][FittingKeys.x_axis] = (
            from_nparray_to_list(fitted_dict[KropffSessionSubKeys.high_tof][FittingKeys.x_axis])
        )
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.high_tof][FittingKeys.y_axis] = (
            from_nparray_to_list(fitted_dict[KropffSessionSubKeys.high_tof][FittingKeys.y_axis])
        )

        # low tof
        cleaned_dict[KropffSessionSubKeys.fitted][KropffTabSelected.low_tof] = {}
        cleaned_dict[KropffSessionSubKeys.fitted][KropffTabSelected.low_tof][FittingKeys.x_axis] = from_nparray_to_list(
            fitted_dict[KropffTabSelected.low_tof][FittingKeys.x_axis]
        )
        cleaned_dict[KropffSessionSubKeys.fitted][KropffTabSelected.low_tof][FittingKeys.y_axis] = from_nparray_to_list(
            fitted_dict[KropffTabSelected.low_tof][FittingKeys.y_axis]
        )

        # bragg peak
        cleaned_dict[KropffSessionSubKeys.fitted][KropffTabSelected.bragg_peak] = {}
        cleaned_dict[KropffSessionSubKeys.fitted][KropffTabSelected.bragg_peak][FittingKeys.x_axis] = (
            from_nparray_to_list(fitted_dict[KropffTabSelected.bragg_peak][FittingKeys.x_axis])
        )
        cleaned_dict[KropffSessionSubKeys.fitted][KropffTabSelected.bragg_peak][FittingKeys.y_axis] = (
            from_nparray_to_list(fitted_dict[KropffTabSelected.bragg_peak][FittingKeys.y_axis])
        )

        # fitting parameters
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.a0] = metadata_for_this_row[
            KropffSessionSubKeys.a0
        ]
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.b0] = metadata_for_this_row[
            KropffSessionSubKeys.b0
        ]
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.ahkl] = metadata_for_this_row[
            KropffSessionSubKeys.ahkl
        ]
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.bhkl] = metadata_for_this_row[
            KropffSessionSubKeys.bhkl
        ]
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.lambda_hkl] = metadata_for_this_row[
            KropffSessionSubKeys.lambda_hkl
        ]
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.tau] = metadata_for_this_row[
            KropffSessionSubKeys.tau
        ]
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.sigma] = metadata_for_this_row[
            KropffSessionSubKeys.sigma
        ]
        cleaned_dict[KropffSessionSubKeys.fitted][KropffSessionSubKeys.bhkl] = metadata_for_this_row[
            KropffSessionSubKeys.bhkl
        ]
        cleaned_dict[KropffSessionSubKeys.fitted][SessionSubKeys.bin_coordinates] = {
            "x0": int(metadata_for_this_row[SessionSubKeys.bin_coordinates]["x0"]),
            "y0": int(metadata_for_this_row[SessionSubKeys.bin_coordinates]["y0"]),
            "x1": int(metadata_for_this_row[SessionSubKeys.bin_coordinates]["x1"]),
            "y1": int(metadata_for_this_row[SessionSubKeys.bin_coordinates]["y1"]),
        }
        cleaned_dict[KropffSessionSubKeys.fitted][SessionSubKeys.bragg_peak_threshold] = metadata_for_this_row[
            SessionSubKeys.bragg_peak_threshold
        ]

        with open(full_output_filename, "w") as json_file:
            json.dump(cleaned_dict, json_file)

    def replace_bragg_peak_row_values(self):
        """replace by median of surrounding pixels"""
        pass

    def display_fitting_parameters(self):
        FittingParametersViewerEditorLauncher(parent=self.parent, grand_parent=self.grand_parent)

    def unlock_all_bragg_peak_rows(self):
        background_color = UNLOCK_ROW_BACKGROUND

        o_table = TableHandler(table_ui=self.parent.ui.bragg_edge_tableWidget)
        nbr_row = o_table.row_count()
        for _row in np.arange(nbr_row):
            o_table.set_background_color_of_row(row=_row, qcolor=background_color)

        o_table_high = TableHandler(table_ui=self.parent.ui.high_lda_tableWidget)
        for _row in np.arange(nbr_row):
            o_table_high.set_background_color_of_row(row=_row, qcolor=background_color)

        o_table_low = TableHandler(table_ui=self.parent.ui.low_lda_tableWidget)
        for _row in np.arange(nbr_row):
            o_table_low.set_background_color_of_row(row=_row, qcolor=background_color)

    def unlock_all_rows_in_table_dictionary(self):
        table_dictionary = self.grand_parent.kropff_table_dictionary
        for _key in table_dictionary.keys():
            table_dictionary[_key]["lock"] = False
        self.grand_parent.kropff_table_dictionary = table_dictionary

    # def bragg_peak_auto_lock_clicked(self):
    #     """if the condition found in the Bragg Edge table are met, the row of all the table will be locked"""
    #
    #     o_table_bragg_peak = TableHandler(table_ui=self.parent.ui.bragg_edge_tableWidget)
    #     o_table_high_tof = TableHandler(table_ui=self.parent.ui.high_lda_tableWidget)
    #     o_table_low_tof = TableHandler(table_ui=self.parent.ui.low_lda_tableWidget)
    #     nbr_row = o_table_bragg_peak.row_count()
    #
    #     table_dictionary = self.grand_parent.kropff_table_dictionary
    #     if self.parent.ui.checkBox.isChecked():
    #
    #         for _row in np.arange(nbr_row):
    #
    #             if self._lets_reject_this_row(row=_row):
    #                 background_color = REJECTED_ROW_BACKGROUND
    #                 table_dictionary[str(_row)]['rejected'] = True
    #
    #             elif self._lets_lock_this_row(row=_row):
    #                 table_dictionary[str(_row)]['rejected'] = False
    #                 background_color = LOCK_ROW_BACKGROUND
    #                 table_dictionary[str(_row)]['lock'] = True
    #
    #             else:
    #                 table_dictionary[str(_row)]['rejected'] = False
    #                 background_color = UNLOCK_ROW_BACKGROUND
    #                 table_dictionary[str(_row)]['lock'] = False
    #
    #             o_table_bragg_peak.set_background_color_of_row(row=_row,
    #                                                            qcolor=background_color)
    #             o_table_high_tof.set_background_color_of_row(row=_row,
    #                                                          qcolor=background_color)
    #             o_table_low_tof.set_background_color_of_row(row=_row,
    #                                                         qcolor=background_color)
    #
    #     else:
    #         for _row in np.arange(nbr_row):
    #
    #             # if table_dictionary[str(_row)]['rejected']:
    #             #     background_color = REJECTED_ROW_BACKGROUND
    #             # else:
    #             background_color = UNLOCK_ROW_BACKGROUND
    #
    #             o_table_bragg_peak.set_background_color_of_row(row=_row,
    #                                                            qcolor=background_color)
    #             o_table_high_tof.set_background_color_of_row(row=_row,
    #                                                          qcolor=background_color)
    #             o_table_low_tof.set_background_color_of_row(row=_row,
    #                                                         qcolor=background_color)
    #
    #         self.unlock_all_rows_in_table_dictionary()

    def _lets_reject_this_row(self, row=0):
        o_table = TableHandler(table_ui=self.parent.ui.high_lda_tableWidget)

        # high lambda
        # if a0 or b0 are nan -> yes, reject this row
        a0_value = o_table.get_item_float_from_cell(row=row, column=FittingKropffHighLambdaColumns.a0)
        if not np.isfinite(a0_value):
            return True

        b0_value = o_table.get_item_float_from_cell(row=row, column=FittingKropffHighLambdaColumns.b0)
        if not np.isfinite(b0_value):
            return True

        # low lambda
        # if ahkl, bhkl are nan -> yes, reject this row
        ahkl_value = o_table.get_item_float_from_cell(row=row, column=FittingKropffLowLambdaColumns.ahkl)
        if not np.isfinite(ahkl_value):
            return True

        bhkl_value = o_table.get_item_float_from_cell(row=row, column=FittingKropffLowLambdaColumns.bhkl)
        if not np.isfinite(bhkl_value):
            return True

        # bragg peak
        # if l_hkl is nan -> yes, reject this row
        l_hkl_value = o_table.get_item_float_from_cell(row=row, column=FittingKropffBraggPeakColumns.l_hkl_value)
        if not np.isfinite(l_hkl_value):
            return True

        # if l_hkl is outside of range defined in settings -> reject this row
        less_than_state = self.parent.kropff_bragg_peak_row_rejections_conditions["l_hkl"]["less_than"]["state"]
        if less_than_state:
            less_than_value = self.parent.kropff_bragg_peak_row_rejections_conditions["l_hkl"]["less_than"]["value"]
            if l_hkl_value < less_than_value:
                return True

        more_than_state = self.parent.kropff_bragg_peak_row_rejections_conditions["l_hkl"]["more_than"]["state"]
        if more_than_state:
            more_than_value = self.parent.kropff_bragg_peak_row_rejections_conditions["l_hkl"]["more_than"]["value"]
            if l_hkl_value > more_than_value:
                return True

        return False

    def _lets_lock_this_row(self, row=0):
        fit_conditions = self.parent.kropff_bragg_peak_good_fit_conditions
        o_table = TableHandler(table_ui=self.parent.ui.bragg_edge_tableWidget)

        o_checking = CheckingFittingConditions(fit_conditions=fit_conditions)
        l_hkl_error = o_table.get_item_float_from_cell(row=row, column=FittingKropffBraggPeakColumns.l_hkl_error)
        t_error = o_table.get_item_float_from_cell(row=row, column=FittingKropffBraggPeakColumns.tau_error)
        sigma_error = o_table.get_item_float_from_cell(row=row, column=FittingKropffBraggPeakColumns.sigma_error)

        return o_checking.is_fitting_ok(l_hkl_error=l_hkl_error, t_error=t_error, sigma_error=sigma_error)

    def check_how_many_fitting_are_locked(self):
        table_dictionary = self.grand_parent.kropff_table_dictionary

        total_number_of_fitting = len(table_dictionary.keys())
        total_number_of_good_fitting = 0
        for _key in table_dictionary.keys():
            lock_state = table_dictionary[_key]["lock"]
            if lock_state:
                total_number_of_good_fitting += 1

        percentage = 100 * (total_number_of_good_fitting / total_number_of_fitting)
        message = (
            f"Percentage of Bragg peak fitted showing uncertainties within the constraint ranges defined:"
            f" {percentage:.2f}%"
        )

        show_status_message(parent=self.parent, message=message, status=StatusMessageStatus.ready)
        QApplication.processEvents()

    def update_summary_table(self):
        table_dictionary = self.grand_parent.kropff_table_dictionary

        list_hkl = []
        list_hkl_error = []

        number_of_fits_locked = 0
        for _key in table_dictionary.keys():
            list_hkl.append(table_dictionary[_key]["lambda_hkl"]["val"])
            list_hkl_error.append(table_dictionary[_key]["lambda_hkl"]["err"])
            if table_dictionary[_key]["lock"]:
                number_of_fits_locked += 1

        # turning None into NaN
        list_hkl_without_none = [_value for _value in list_hkl if _value is not None]
        list_hkl_error_without_none = [_value for _value in list_hkl_error if _value is not None]

        number_of_fittings = len(list_hkl_without_none)
        number_of_fittings_with_error = len(list_hkl_error_without_none)
        total_number_of_bins = len(list_hkl)

        hkl_value_mean = np.mean(list_hkl_without_none)
        hkl_value_median = np.median(list_hkl_without_none)
        hkl_value_std = np.std(list_hkl_without_none)
        hkl_value_percentage_with_fit = 100 * (number_of_fittings / total_number_of_bins)

        hkl_error_value_mean = np.mean(list_hkl_error_without_none)
        hkl_error_value_median = np.median(list_hkl_error_without_none)
        hkl_error_value_std = np.std(list_hkl_error_without_none)
        hkl_error_value_percentage_with_fit = 100 * (number_of_fittings_with_error / total_number_of_bins)

        percentage_of_fits_locked = 100 * (number_of_fits_locked / total_number_of_bins)

        o_table = TableHandler(table_ui=self.parent.ui.kropff_summary_tableWidget)
        o_table.insert_item(row=0, column=1, value=hkl_value_mean, format_str="{:0.4f}", editable=False)
        o_table.insert_item(
            row=1,
            column=1,
            value=hkl_value_median,
            format_str="{:0.4f}",
            editable=False,
        )
        o_table.insert_item(row=2, column=1, value=hkl_value_std, format_str="{:0.4f}", editable=False)
        o_table.insert_item(
            row=3,
            column=1,
            value=hkl_value_percentage_with_fit,
            format_str="{:0.2f}",
            editable=False,
        )
        o_table.insert_item(
            row=4,
            column=1,
            value=percentage_of_fits_locked,
            format_str="{:0.2f}",
            editable=False,
        )

        o_table.insert_item(
            row=0,
            column=2,
            value=hkl_error_value_mean,
            format_str="{:0.4f}",
            editable=False,
        )
        o_table.insert_item(
            row=1,
            column=2,
            value=hkl_error_value_median,
            format_str="{:0.4f}",
            editable=False,
        )
        o_table.insert_item(
            row=2,
            column=2,
            value=hkl_error_value_std,
            format_str="{:0.4f}",
            editable=False,
        )
        o_table.insert_item(
            row=3,
            column=2,
            value=hkl_error_value_percentage_with_fit,
            format_str="{:0.2f}",
            editable=False,
        )

    # top left view mouse events
    def mouse_clicked_in_top_left_image_view(self, mouse_click_event):
        image_pos = self.parent.image_view_item.mapFromScene(mouse_click_event.scenePos())

        # if user click within a BIN, select that bin in all the tables (this will automatically highlight it
        [_, top_left_x, top_left_y, _, _, binning_size] = self.grand_parent.session_dict[DataType.bin][
            SessionSubKeys.roi
        ]

        # top_left_corner_coordinates = self.grand_parent.binning_line_view['pos'][0]
        # top_left_x = top_left_corner_coordinates[0]
        # top_left_y = top_left_corner_coordinates[1]
        #
        # binning_size = self.grand_parent.binning_roi[-1]
        x = int(image_pos.x())
        y = int(image_pos.y())

        if x < top_left_x:
            return

        if y < top_left_y:
            return

        o_get = KropffGet(parent=self.parent)
        tab_selected = o_get.kropff_tab_selected()

        if tab_selected not in [
            KropffTabSelected.low_tof,
            KropffTabSelected.high_tof,
            KropffTabSelected.bragg_peak,
        ]:
            # no need to update table as we are showing a tab without tables
            return

        table_ui = o_get.kropff_table_ui_selected()
        nbr_bin_y_direction = self.grand_parent.fitting_selection["nbr_row"]
        nbr_bin_x_direction = self.grand_parent.fitting_selection["nbr_column"]

        if x > (top_left_x + (nbr_bin_x_direction + 1) * binning_size):
            return

        if y > (top_left_y + (nbr_bin_y_direction) * binning_size):
            return

        bin_x_index = int((x - top_left_x) / binning_size) + 1
        bin_y_index = int((y - top_left_y) / binning_size) + 1

        row_to_select = int(bin_y_index + (bin_x_index - 1) * nbr_bin_y_direction - 1)
        o_table = TableHandler(table_ui=table_ui)
        o_table.select_row(row_to_select)

    def mouse_moved_in_top_left_image_view(self, evt):
        pos = evt[0]

        width = self.grand_parent.data_metadata[DataType.normalized]["size"]["width"]
        height = self.grand_parent.data_metadata[DataType.normalized]["size"]["height"]

        if self.parent.image_view_item.sceneBoundingRect().contains(pos):
            image_pos = self.parent.image_view_item.mapFromScene(pos)
            x = int(image_pos.x())
            y = int(image_pos.y())
            binning_size = self.grand_parent.session_dict[DataType.bin][SessionSubKeys.roi][-1]
            top_left_corner_coordinates = self.grand_parent.binning_line_view["pos"][0]
            top_left_x = top_left_corner_coordinates[0]
            top_left_y = top_left_corner_coordinates[1]
            bin_x_index = int((x - top_left_x) / binning_size) + 1
            bin_y_index = int((y - top_left_y) / binning_size) + 1
            nbr_bin_y_direction = self.grand_parent.fitting_selection["nbr_row"]
            row_to_select = int(bin_y_index + (bin_x_index - 1) * nbr_bin_y_direction - 1) + 1

            if (x >= 0) and (x < width) and (y >= 0) and (y < height):
                self.parent.image_view_vline.setPos(x)
                self.parent.image_view_hline.setPos(y)

                self.parent.ui.kropff_pos_x_value.setText(f"{x}")
                self.parent.ui.kropff_pos_y_value.setText(f"{y}")

                # only if we are inside the bin selection
                bin_list = self.grand_parent.session_dict[DataType.bin][SessionSubKeys.roi]
                left = bin_list[1]
                top = bin_list[2]
                roi_width = bin_list[3]
                roi_height = bin_list[4]
                bin_size = bin_list[5]

                if (
                    (x >= left)
                    and (x <= (left + roi_width - bin_size))
                    and (y >= top)
                    and (y <= (top + roi_height - bin_size))
                ):
                    # inside the ROI region
                    self.parent.ui.kropff_bin_x_value.setText(f"{bin_x_index}")
                    self.parent.ui.kropff_bin_y_value.setText(f"{bin_y_index}")
                    self.parent.ui.kropff_bin_nbr_value.setText(f"{row_to_select}")
                else:
                    # outside the ROI region
                    self.parent.ui.kropff_bin_x_value.setText("N/A")
                    self.parent.ui.kropff_bin_y_value.setText("N/A")
                    self.parent.ui.kropff_bin_nbr_value.setText("N/A")

            else:
                self.parent.ui.kropff_pos_x_value.setText("N/A")
                self.parent.ui.kropff_pos_y_value.setText("N/A")
                self.parent.ui.kropff_bin_x_value.setText("N/A")
                self.parent.ui.kropff_bin_y_value.setText("N/A")
                self.parent.ui.kropff_bin_nbr_value.setText("N/A")

        else:
            self.parent.ui.kropff_pos_x_value.setText("N/A")
            self.parent.ui.kropff_pos_y_value.setText("N/A")
            self.parent.ui.kropff_bin_x_value.setText("N/A")
            self.parent.ui.kropff_bin_y_value.setText("N/A")
            self.parent.ui.kropff_bin_nbr_value.setText("N/A")

    def change_initial_guess_lambda_hkl_widgets(self):
        state_fix = self.parent.ui.lambda_hkl_fix_radioButton.isChecked()

        self.parent.ui.lambda_hkl_fix_lineEdit.setEnabled(state_fix)
        self.parent.ui.lambda_hkl_from_lineEdit.setEnabled(not state_fix)
        self.parent.ui.to_label.setEnabled(not state_fix)
        self.parent.ui.lambda_hkl_to_lineEdit.setEnabled(not state_fix)
        self.parent.ui.step_label.setEnabled(not state_fix)
        self.parent.ui.lambda_hkl_step_lineEdit.setEnabled(not state_fix)

    def change_initial_guess_tau_widgets(self):
        state_fix = self.parent.ui.tau_fix_radioButton.isChecked()

        self.parent.ui.tau_fix_lineEdit.setEnabled(state_fix)
        self.parent.ui.tau_from_lineEdit.setEnabled(not state_fix)
        self.parent.ui.to_label_1.setEnabled(not state_fix)
        self.parent.ui.tau_to_lineEdit.setEnabled(not state_fix)
        self.parent.ui.step_label_1.setEnabled(not state_fix)
        self.parent.ui.tau_step_lineEdit.setEnabled(not state_fix)

    def change_initial_guess_sigma_widgets(self):
        state_fix = self.parent.ui.sigma_fix_radioButton.isChecked()

        self.parent.ui.sigma_fix_lineEdit.setEnabled(state_fix)
        self.parent.ui.sigma_from_lineEdit.setEnabled(not state_fix)
        self.parent.ui.to_label_2.setEnabled(not state_fix)
        self.parent.ui.sigma_to_lineEdit.setEnabled(not state_fix)
        self.parent.ui.step_label_2.setEnabled(not state_fix)
        self.parent.ui.sigma_step_lineEdit.setEnabled(not state_fix)



from qtpy.QtGui import QGuiApplication
from qtpy.QtWidgets import QApplication


class StatusMessageStatus:
    ready = "QStatusBar{padding-left:8px;color:green;font-weight:normal;}"
    working = "QStatusBar{padding-left:8px;color:blue;font-weight:normal;}"
    error = "QStatusBar{padding-left:8px;;color:red;font-weight:bold;}"
    warning = "QStatusBar{padding-left:8px;color:purple;font-weight:normal;}"


def show_status_message(parent=None, message="", status=StatusMessageStatus.ready, duration_s=None):
    parent.ui.statusbar.setStyleSheet(status)
    if duration_s:
        parent.ui.statusbar.showMessage(message, duration_s * 1000)
    else:
        parent.ui.statusbar.showMessage(message)
    QGuiApplication.processEvents()
    parent.ui.repaint()
    QApplication.processEvents()



#!/usr/bin/env python
"""
Preview full bin axis
"""

from qtpy.QtWidgets import QDialog

from ibeatles import load_ui
from ibeatles.tools import ANGSTROMS, MICRO
from ibeatles.tools.tof_bin import TO_ANGSTROMS_UNITS, TO_MICROS_UNITS
from ibeatles.tools.tof_bin.utilities.get import Get
from ibeatles.tools.utilities import TimeSpectraKeys
from ibeatles.utilities.table_handler import TableHandler


class PreviewFullBinAxis(QDialog):
    def __init__(self, parent=None):
        self.parent = parent

        QDialog.__init__(self, parent=parent)
        self.ui = load_ui("ui_preview_full_bin_axis.ui", baseinstance=self)
        self.setWindowTitle("Full Bin Axis Requested")

        self.update_widgets()

    def update_widgets(self):
        o_get = Get(parent=self.parent)

        # top labels
        bin_type = o_get.bin_auto_mode()
        self.ui.bin_type_label.setText(bin_type)

        bin_axis = o_get.current_bin_tab_working_axis()
        self.ui.axis_label.setText(bin_axis)

        if bin_axis == TimeSpectraKeys.file_index_array:
            units = "N/A"
            str_format = "{}"
            coeff = 1
        elif bin_axis == TimeSpectraKeys.tof_array:
            units = f"{MICRO}s"
            str_format = "{:.2f}"
            coeff = TO_MICROS_UNITS
        else:
            units = f"{ANGSTROMS}"
            str_format = "{:.3f}"
            coeff = TO_ANGSTROMS_UNITS

        self.ui.units_label.setText(units)

        # table
        full_bin_axis_requested = self.parent.full_bin_axis_requested
        o_table = TableHandler(table_ui=self.ui.tableWidget)
        for _row, _left_bin_value in enumerate(full_bin_axis_requested[:-1]):
            o_table.insert_empty_row(row=_row)
            o_table.insert_item(
                row=_row,
                column=0,
                value=_left_bin_value * coeff,
                format_str=str_format,
                editable=False,
            )
            o_table.insert_item(
                row=_row,
                column=1,
                value=full_bin_axis_requested[_row + 1] * coeff,
                format_str=str_format,
                editable=False,
            )



#!/usr/bin/env python
"""Uncertainty calculations for TPX1 MCP detector-corrected data.

This module computes uncertainty estimates for detector-efficiency-corrected
neutron imaging data. It takes the corrected TIFF stack and metadata files
as input, and outputs counts with uncertainty for a specified ROI.

The uncertainty is derived from the detector correction formula:
    T_i = F_i / ((1 - P_i) * s_i)

Where:
    - T_i: corrected counts (what the user has)
    - F_i: raw counts (derived internally)
    - P_i: cumulative pixel occupancy (derived internally)
    - s_i: shutter normalization ratio (derived from shutter counts file)

The variance formula (validated via Monte Carlo):
    Var(T_i) = T_i / ((1 - P_i) * s_i)

Usage
-----
>>> from ibeatles.core.detector import compute_counts_with_uncertainty
>>> result = compute_counts_with_uncertainty(
...     tiff_stack=corrected_images,  # shape: (n_frames, height, width)
...     tof_array=tof_values,         # shape: (n_frames,)
...     shutter_counts_file="path/to/ShutterCount.txt",
...     roi_mask=roi,                 # shape: (height, width), boolean
... )
>>> result.to_csv("output.csv", index=False)

References
----------
- Validation notebook: notebooks/uncertainty_validation.ipynb
- Reference implementation: NeutronImagingScripts/neutronimaging/detector_correction.py
"""

from pathlib import Path
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd
from loguru import logger


def load_shutter_counts(file_path: Union[str, Path]) -> pd.DataFrame:
    """Load and parse shutter counts file.

    Parameters
    ----------
    file_path : str or Path
        Path to the shutter counts file (tab-separated, .txt extension).
        Expected format: shutter_index<TAB>shutter_counts

    Returns
    -------
    pd.DataFrame
        DataFrame with columns:
        - shutter_index: int
        - shutter_counts: int (total neutron counts for this shutter period)
        - shutter_n_ratio: float (shutter_counts / shutter_counts[0])

    Notes
    -----
    Only rows with shutter_counts > 0 are included.
    The shutter_n_ratio is derived as counts[i] / counts[0].
    """
    file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(f"Shutter counts file not found: {file_path}")

    df = pd.read_csv(
        file_path,
        sep="\t",
        names=["shutter_index", "shutter_counts"],
    )
    # Filter out zero-count entries
    df = df[df["shutter_counts"] > 0].copy()

    if len(df) == 0:
        raise ValueError(f"No valid shutter counts found in {file_path}")

    # Derive the normalization ratio
    df["shutter_n_ratio"] = df["shutter_counts"] / df["shutter_counts"].values[0]

    logger.info(
        f"Loaded shutter counts: {len(df)} shutter periods, "
        f"counts range [{df['shutter_counts'].min()}, {df['shutter_counts'].max()}]"
    )

    return df.reset_index(drop=True)


def _recover_raw_counts(
    corrected: np.ndarray,
    shutter_counts: int,
    shutter_n_ratio: Union[np.ndarray, float],
) -> Tuple[np.ndarray, np.ndarray]:
    """Recover raw counts F_i from corrected counts T_i.

    Uses the inversion formula derived from T_i = F_i / ((1 - P_i) * s_i):
        F_i = T_i * s_i * (N - S_{i-1}) / (N + T_i * s_i)

    Where S_{i-1} = cumsum of F_j for j < i.

    Parameters
    ----------
    corrected : np.ndarray
        Corrected counts T_i, shape (n_frames, height, width).
    shutter_counts : int
        Total shutter counts N for this shutter period.
    shutter_n_ratio : np.ndarray or float
        Shutter normalization ratio s_i. If array, shape (n_frames,).

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        (raw_counts, occupancy) - F_i and P_i arrays with same shape as input.
    """
    n_frames = corrected.shape[0]
    N = float(shutter_counts)

    # Handle scalar vs array shutter ratio
    if np.isscalar(shutter_n_ratio):
        s = np.full(n_frames, shutter_n_ratio)
    else:
        s = np.asarray(shutter_n_ratio)

    # Initialize output arrays
    raw = np.zeros_like(corrected, dtype=np.float64)
    occupancy = np.zeros_like(corrected, dtype=np.float64)

    # Iteratively compute F_i and P_i
    cumsum_raw = np.zeros(corrected.shape[1:], dtype=np.float64)  # S_{i-1}

    for i in range(n_frames):
        T_i = corrected[i].astype(np.float64)
        s_i = s[i]

        # F_i = T_i * s_i * (N - S_{i-1}) / (N + T_i * s_i)
        numerator = T_i * s_i * (N - cumsum_raw)
        denominator = N + T_i * s_i

        # Avoid division by zero and negative counts
        # (numerator can be negative if cumsum_raw exceeds N due to numerical issues)
        with np.errstate(divide="ignore", invalid="ignore"):
            F_i = np.where((denominator > 0) & (numerator >= 0), numerator / denominator, 0.0)

        raw[i] = F_i

        # Update cumsum for next iteration
        cumsum_raw = cumsum_raw + F_i

        # P_i = cumsum(F) / N (including current frame)
        occupancy[i] = cumsum_raw / N

    return raw, occupancy


def _compute_variance(
    corrected: np.ndarray,
    occupancy: np.ndarray,
    shutter_n_ratio: Union[np.ndarray, float],
) -> np.ndarray:
    """Compute variance of corrected counts.

    Var(T_i) = T_i / ((1 - P_i) * s_i)

    Parameters
    ----------
    corrected : np.ndarray
        Corrected counts T_i.
    occupancy : np.ndarray
        Pixel occupancy probability P_i.
    shutter_n_ratio : np.ndarray or float
        Shutter normalization ratio s_i.

    Returns
    -------
    np.ndarray
        Variance array with same shape as input.
    """
    # Broadcast shutter ratio if needed
    if np.isscalar(shutter_n_ratio):
        s = shutter_n_ratio
    else:
        s = np.asarray(shutter_n_ratio)
        if s.ndim == 1:
            # Shape (n_frames,) -> (n_frames, 1, 1) for broadcasting
            s = s[:, np.newaxis, np.newaxis]

    # Compute variance
    denominator = (1.0 - occupancy) * s

    with np.errstate(divide="ignore", invalid="ignore"):
        variance = np.where(denominator > 0, corrected / denominator, 0.0)

    # Ensure non-negative
    variance = np.maximum(variance, 0.0)

    return variance


def compute_counts_with_uncertainty(
    tiff_stack: np.ndarray,
    tof_array: np.ndarray,
    shutter_counts_file: Union[str, Path],
    roi_mask: Optional[np.ndarray] = None,
    shutter_index: int = 0,
) -> pd.DataFrame:
    """Compute counts with uncertainty for a region of interest.

    This is the main entry point for uncertainty calculation. It takes
    the detector-efficiency-corrected TIFF stack and metadata files,
    computes the uncertainty, and returns a DataFrame ready for export.

    Parameters
    ----------
    tiff_stack : np.ndarray
        Detector-efficiency-corrected image stack, shape (n_frames, height, width).
        These are the T_i values from the correction formula.
    tof_array : np.ndarray
        Time-of-flight values for each frame, shape (n_frames,).
        Can be obtained from load_time_spectra()["tof_array"].
    shutter_counts_file : str or Path
        Path to the shutter counts file.
    roi_mask : np.ndarray, optional
        Boolean mask defining the ROI, shape (height, width).
        If None, uses all pixels.
    shutter_index : int, optional
        Which shutter period to use (default 0, the first).
        For single-shutter acquisitions, use 0.

    Returns
    -------
    pd.DataFrame
        DataFrame with columns:
        - tof: Time-of-flight values
        - counts: Summed counts in ROI
        - uncertainty: Uncertainty (standard deviation) of summed counts

    Examples
    --------
    >>> # Load data using existing iBeatles functions
    >>> data = load_data_from_folder("/path/to/corrected/tiffs")
    >>> spectra = load_time_spectra(spectra_file, distance, offset)
    >>>
    >>> # Define ROI (e.g., 100x100 box in center)
    >>> roi = np.zeros((512, 512), dtype=bool)
    >>> roi[200:300, 200:300] = True
    >>>
    >>> # Compute uncertainty
    >>> result = compute_counts_with_uncertainty(
    ...     tiff_stack=data["data"],
    ...     tof_array=spectra["tof_array"],
    ...     shutter_counts_file="/path/to/ShutterCount.txt",
    ...     roi_mask=roi,
    ... )
    >>>
    >>> # Export to CSV
    >>> result.to_csv("counts_with_uncertainty.csv", index=False)
    """
    # Validate inputs
    tiff_stack = np.asarray(tiff_stack, dtype=np.float64)
    tof_array = np.asarray(tof_array)

    if tiff_stack.ndim != 3:
        raise ValueError(f"tiff_stack must be 3D (n_frames, height, width), got shape {tiff_stack.shape}")

    n_frames, height, width = tiff_stack.shape

    if n_frames == 0:
        raise ValueError("tiff_stack must contain at least one frame")

    if len(tof_array) != n_frames:
        raise ValueError(f"tof_array length ({len(tof_array)}) must match number of frames ({n_frames})")

    # Load shutter metadata
    shutter_df = load_shutter_counts(shutter_counts_file)

    if shutter_index < 0 or shutter_index >= len(shutter_df):
        raise ValueError(f"shutter_index {shutter_index} out of range (must be 0 to {len(shutter_df) - 1})")

    shutter_counts = int(shutter_df.loc[shutter_index, "shutter_counts"])
    shutter_n_ratio = float(shutter_df.loc[shutter_index, "shutter_n_ratio"])

    if shutter_counts <= 0:
        raise ValueError(f"shutter_counts must be positive, got {shutter_counts}")
    if shutter_n_ratio <= 0:
        raise ValueError(f"shutter_n_ratio must be positive, got {shutter_n_ratio}")

    logger.info(f"Using shutter period {shutter_index}: counts={shutter_counts}, ratio={shutter_n_ratio:.4f}")

    # Create ROI mask if not provided
    if roi_mask is None:
        roi_mask = np.ones((height, width), dtype=bool)
        logger.info("No ROI mask provided, using all pixels")
    else:
        roi_mask = np.asarray(roi_mask, dtype=bool)
        if roi_mask.shape != (height, width):
            raise ValueError(f"roi_mask shape {roi_mask.shape} must match image shape ({height}, {width})")
        n_pixels = np.sum(roi_mask)
        if n_pixels == 0:
            raise ValueError("roi_mask must contain at least one pixel")
        logger.info(f"ROI contains {n_pixels} pixels")

    # Recover raw counts and compute occupancy
    logger.info("Computing occupancy from corrected counts...")
    _, occupancy = _recover_raw_counts(tiff_stack, shutter_counts, shutter_n_ratio)

    # Compute variance
    logger.info("Computing variance...")
    variance = _compute_variance(tiff_stack, occupancy, shutter_n_ratio)

    # Aggregate over ROI (vectorized for performance)
    logger.info("Aggregating over ROI...")
    counts_per_frame = np.sum(tiff_stack[:, roi_mask], axis=1)
    variance_per_frame = np.sum(variance[:, roi_mask], axis=1)

    # Standard deviation = sqrt(variance)
    uncertainty_per_frame = np.sqrt(variance_per_frame)

    # Build output DataFrame
    result = pd.DataFrame(
        {
            "tof": tof_array,
            "counts": counts_per_frame,
            "uncertainty": uncertainty_per_frame,
        }
    )

    logger.info(f"Computed uncertainty for {n_frames} frames")

    return result



import glob
import json
import os
import shutil
from collections import Counter
from datetime import datetime
from pathlib import Path

import numpy as np
from astropy.io import fits
from PIL import Image
from qtpy.QtWidgets import QFileDialog

TIME_SPECTRA_NAME_FORMAT = "*_Spectra.txt"


class FileHandler:
    @classmethod
    def get_list_of_folders(cls, top_folder):
        full_list = glob.glob(top_folder + "/*")
        full_list.sort()
        list_folders = []
        for _entry in full_list:
            if os.path.isdir(_entry):
                list_folders.append(_entry)
        return list_folders

    @classmethod
    def get_list_of_all_subfolders(cls, top_folder):
        tuple_folders = os.walk(top_folder)
        return [x[0] for x in tuple_folders]

    @classmethod
    def get_list_of_tif(cls, folder):
        full_list = glob.glob(folder + "/*.tif*")
        full_list.sort()
        return full_list

    @classmethod
    def get_parent_folder(cls, full_folder):
        folder_parts = Path(full_folder).parts
        return folder_parts[-2]

    @classmethod
    def get_parent_path(cls, full_path):
        full_path = Path(full_path)
        return full_path.parent

    @classmethod
    def get_base_filename(cls, full_file_name):
        return Path(full_file_name).name

    @staticmethod
    def get_file_extension(file_name):
        ext = Path(file_name).suffix
        return ext

    @classmethod
    def retrieve_ascii_contain(cls, full_file_name):
        file_contain = []
        with open(full_file_name) as f:
            file_contain = f.read()
        return file_contain

    @classmethod
    def cleanup_list_of_files(cls, list_of_files=[], base_number=5):
        """Will only keep the files that have the same number of character as
        the first n files"""
        if len(list_of_files) == 0:
            return []

        len_base_files = []
        for _file in list_of_files[0:base_number]:
            len_base_files.append(len(_file))

        # make sure all the length of the base number files match
        set_len_base_files = set(len_base_files)
        if len(set_len_base_files) > 1:
            raise ValueError("Format Input File Do Not Match!")

        len_file = len_base_files[0]
        final_list = []
        for _file in list_of_files:
            if len(_file) == len_file:
                final_list.append(_file)

        return final_list

    @classmethod
    def make_fits(cls, data=[], filename=""):
        fits.writeto(filename, data, overwrite=True)

    @classmethod
    def make_tiff(cls, data=[], filename=""):
        new_image = Image.fromarray(data)
        new_image.save(filename)

    @classmethod
    def make_ascii_file(cls, metadata=[], data=[], output_file_name="", sep=","):
        f = open(output_file_name, "w")
        for _meta in metadata:
            _line = _meta + "\n"
            f.write(_line)

        if len(np.shape(data)) > 1:
            for _data in data:
                _str_data = [str(_value) for _value in _data]
                _line = sep.join(_str_data) + "\n"
                f.write(_line)
        else:
            _str_data = [str(_value) + "\n" for _value in data]
            for _data in _str_data:
                f.write(_data)

        f.close()

    def make_json_file(data_dict: dict = None, output_file_name: str = None):
        with open(output_file_name, "w") as json_file:
            json.dump(data_dict, json_file)

    @staticmethod
    def make_or_reset_folder(folder_name):
        if os.path.exists(folder_name):
            shutil.rmtree(folder_name)
        os.makedirs(folder_name)

    @staticmethod
    def make_folder(folder_name):
        os.makedirs(folder_name)

    @staticmethod
    def make_or_append_date_time_to_folder(folder_name):
        if os.path.exists(folder_name):
            _current_date_time = FileHandler.get_current_timestamp()
            folder_name += f"_{_current_date_time}"
        os.makedirs(folder_name)
        return folder_name

    @staticmethod
    def get_current_timestamp():
        """Convert the unix time stamp into a human-readable time format

        Format return will look like  "y2018_m01_d29_h10_mn30"
        """
        now = datetime.now()
        return now.strftime("y%Y_m%m_d%d_h%H_mn%M")


def retrieve_timestamp_file_name(folder):
    time_spectra = glob.glob(folder + "/" + TIME_SPECTRA_NAME_FORMAT)
    if time_spectra and os.path.exists(time_spectra[0]):
        return f"{time_spectra[0]}"

    else:
        return ""


def read_ascii(filename=""):
    """return contain of an ascii file"""
    with open(filename, "r") as f:
        text = f.read()
    return text


def write_ascii(text="", filename=""):
    with open(filename, "w") as f:
        f.write(text)


def get_current_timestamp():
    """Convert the unix time stamp into a human-readable time format

    Format return will look like  "y2018_m01_d29_h10_mn30"
    """
    now = datetime.now()
    return now.strftime("y%Y_m%m_d%d_h%H_mn%M")


def create_full_export_file_name(base_name, ext):
    """
    Create the name of the file to export all tabs
    """
    file_name = f"{base_name}_{get_current_timestamp()}.{ext}"
    if os.path.exists(file_name):
        file_name = f"{base_name}_{get_current_timestamp()}_{get_current_timestamp()}.{ext}"

    return file_name


def select_folder(start_folder="./"):
    return str(
        QFileDialog.getExistingDirectory(
            caption="Select output folder",
            directory=start_folder,
            options=QFileDialog.ShowDirsOnly,
        )
    )


def get_list_of_most_dominant_extension_from_folder(folder="", files=None):
    """
    This will return the list of files from the most dominant file extension found in the folder
    as well as the most dominant extension used
    """

    if folder:
        list_of_input_files = glob.glob(os.path.join(folder, "*"))
    else:
        list_of_input_files = files

    list_of_input_files.sort()
    list_of_base_name = [os.path.basename(_file) for _file in list_of_input_files]

    # work with the largest common file extension from the folder selected

    counter_extension = Counter()
    for _file in list_of_base_name:
        [_base, _ext] = os.path.splitext(_file)
        counter_extension[_ext] += 1

    dominand_extension = ""
    dominand_number = 0
    for _key in counter_extension.keys():
        if counter_extension[_key] > dominand_number:
            dominand_extension = _key
            dominand_number = counter_extension[_key]

    list_of_input_files = glob.glob(os.path.join(folder, "*" + dominand_extension))
    list_of_input_files.sort()

    list_of_input_files = [os.path.abspath(_file) for _file in list_of_input_files]

    return [list_of_input_files, dominand_extension]



#!/usr/bin/env python
"""
LoadPreviousSessionLauncherMultipleChoice class
"""

import os

from qtpy.QtWidgets import QDialog

from ibeatles import DataType, load_ui
from ibeatles.session.session_handler import SessionHandler


class LoadPreviousSessionLauncherMultipleChoice(QDialog):
    def __init__(self, parent=None, config=None, list_tabs_to_load=None):
        self.parent = parent
        QDialog.__init__(self, parent=parent)
        ui_full_path = os.path.join(
            os.path.dirname(__file__),
            os.path.join("ui", "ui_load_previous_session_multiple_choice.ui"),
        )
        self.ui = load_ui(ui_full_path, baseinstance=self)
        self.setWindowTitle("Select Tabs to Load?")
        self.ui.ok_pushButton.setFocus(True)

        self.init_widgets(list_tabs_to_load=list_tabs_to_load)
        self.check_box_clicked()

    def init_widgets(self, list_tabs_to_load=None):
        if DataType.sample in list_tabs_to_load:
            self.ui.sample_and_ob_checkBox.setEnabled(True)
            self.ui.sample_and_ob_checkBox.setChecked(True)
        if DataType.normalized in list_tabs_to_load:
            self.ui.normalized_checkBox.setEnabled(True)
            self.ui.normalized_checkBox.setChecked(True)
        if DataType.bin in list_tabs_to_load:
            self.ui.bin_checkBox.setEnabled(True)
            self.ui.bin_checkBox.setChecked(True)
        if DataType.fitting in list_tabs_to_load:
            self.ui.bin_checkBox.setChecked(True)
            self.ui.fitting_checkBox.setEnabled(True)
            self.ui.fitting_checkBox.setChecked(True)

    def check_box_clicked(self):
        is_fitting_checked = self.ui.fitting_checkBox.isChecked()
        if is_fitting_checked:
            self.ui.bin_checkBox.setEnabled(False)
            self.ui.bin_checkBox.setChecked(True)
        else:
            self.ui.bin_checkBox.setEnabled(True)

    def _get_list_tabs_to_load(self):
        list_tabs_to_load = []
        if self.ui.sample_and_ob_checkBox.isChecked():
            list_tabs_to_load.append(DataType.sample)
        if self.ui.normalized_checkBox.isChecked():
            list_tabs_to_load.append(DataType.normalized)
        if self.ui.bin_checkBox.isChecked():
            list_tabs_to_load.append(DataType.bin)
        if self.ui.fitting_checkBox.isChecked():
            list_tabs_to_load.append(DataType.fitting)
        return list_tabs_to_load

    def ok_clicked(self):
        self.close()
        o_session = SessionHandler(parent=self.parent)
        o_session.load_to_ui(tabs_to_load=self._get_list_tabs_to_load())
        self.parent.loading_from_config = False

    def cancel_clicked(self):
        self.close()



#!/usr/bin/env python
"""
KropffBraggPeakThresholdCalculator class for handling the automatic Bragg peak threshold calculator.
"""

import numpy as np
from loguru import logger

from ibeatles import DataType
from ibeatles.fitting import FittingTabSelected
from ibeatles.fitting.kropff import SessionSubKeys as KropffSessionSubKeys
from ibeatles.fitting.kropff.kropff_automatic_threshold_algorithms import Algorithms
from ibeatles.utilities.table_handler import TableHandler


class KropffBraggPeakThresholdCalculator:
    def __init__(self, parent=None, grand_parent=None):
        self.parent = parent
        self.grand_parent = grand_parent

    def run_automatic_mode(self):
        logger.info("Automatic Bragg peak threshold calculator")
        kropff_table_dictionary = self.grand_parent.kropff_table_dictionary
        algorithm_selected = self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
            KropffSessionSubKeys.automatic_bragg_peak_threshold_algorithm
        ]
        logger.info(f"-> algorithm selected: {algorithm_selected}")
        progress_bar_ui = self.parent.eventProgress

        o_algo = Algorithms(
            kropff_table_dictionary=kropff_table_dictionary,
            algorithm_selected=algorithm_selected,
            progress_bar_ui=progress_bar_ui,
        )

        list_of_threshold_calculated = o_algo.get_peak_value_array(algorithm_selected)
        logger.info(f"-> list of threshold found: {list_of_threshold_calculated}")

        threshold_width = int(
            self.grand_parent.session_dict[DataType.fitting][FittingTabSelected.kropff][
                KropffSessionSubKeys.automatic_fitting_threshold_width
            ]
        )

        for _row_index, _row in enumerate(kropff_table_dictionary.keys()):
            x_axis = kropff_table_dictionary[_row]["xaxis"]
            left_index = list_of_threshold_calculated[_row_index] - threshold_width
            right_index = list_of_threshold_calculated[_row_index] + threshold_width
            if right_index >= len(x_axis):
                right_index = len(x_axis) - 1
            kropff_table_dictionary[_row]["bragg peak threshold"]["left"] = x_axis[left_index]
            kropff_table_dictionary[_row]["bragg peak threshold"]["right"] = x_axis[right_index]

        self.grand_parent.kropff_table_dictionary = kropff_table_dictionary

    def save_all_profiles(self, force=False):
        o_table = TableHandler(table_ui=self.parent.ui.high_lda_tableWidget)
        nbr_row = o_table.row_count()
        table_dictionary = self.grand_parent.kropff_table_dictionary
        data_2d = self.grand_parent.data_metadata["normalized"]["data"]

        # index of selection in bragg edge plot
        [left_index, right_index] = self.grand_parent.fitting_bragg_edge_linear_selection

        run_calculation = False
        for _row in np.arange(nbr_row):
            _bin_entry = table_dictionary[str(_row)]

            if force:
                run_calculation = True
            elif _bin_entry["yaxis"] is None:
                run_calculation = True

            if run_calculation:
                _bin_x0 = _bin_entry["bin_coordinates"]["x0"]
                _bin_x1 = _bin_entry["bin_coordinates"]["x1"]
                _bin_y0 = _bin_entry["bin_coordinates"]["y0"]
                _bin_y1 = _bin_entry["bin_coordinates"]["y1"]

                yaxis = data_2d[
                    left_index:right_index,
                    _bin_x0:_bin_x1,
                    _bin_y0:_bin_y1,
                ]  # noqa: E124
                yaxis = np.nanmean(yaxis, axis=1)
                yaxis = np.array(np.nanmean(yaxis, axis=1), dtype=float)
                _bin_entry["yaxis"] = yaxis
                self.grand_parent.kropff_table_dictionary[str(_row)] = _bin_entry

                # index of selection in bragg edge plot
                [left_index, right_index] = self.grand_parent.fitting_bragg_edge_linear_selection
                full_x_axis = self.parent.bragg_edge_data["x_axis"]
                xaxis = np.array(full_x_axis[left_index:right_index], dtype=float)
                _bin_entry["xaxis"] = xaxis



"""Command line interface for iBeatles."""

#!/usr/bin/env python3
import argparse
import json
import logging
from pathlib import Path
from typing import Any, Dict, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from scipy.ndimage import gaussian_filter1d

from ibeatles.core.config import IBeatlesUserConfig
from ibeatles.core.fitting.binning import get_bin_coordinates, get_bin_transmission
from ibeatles.core.fitting.kropff.fitting import fit_bragg_edge_single_pass
from ibeatles.core.io.data_loading import (
    get_time_spectra_filename,
    load_data_from_folder,
    load_time_spectra,
)
from ibeatles.core.material import get_initial_bragg_edge_lambda
from ibeatles.core.processing.normalization import normalize_data
from ibeatles.core.strain.export import (
    generate_output_filename,
    save_analysis_results,
    save_fitting_grid,
    save_strain_map,
)
from ibeatles.core.strain.mapping import calculate_strain_mapping
from ibeatles.core.strain.visualization import (
    plot_fitting_results_grid,
    plot_strain_map_overlay,
)


def setup_logging(log_file: Optional[Path] = None, log_level: int = logging.INFO) -> logging.Logger:
    """
    Set up logging for the application.

    Parameters
    ----------
    log_file : Path, optional
        Path to the log file. If not provided, logs will be saved in the current working directory.
    log_level : int, optional
        Logging level. Defaults to logging.INFO.

    Returns
    -------
    logging.Logger
        Configured logger instance.
    """
    if log_file is None:
        log_file = Path.cwd() / "ibeatles_cli.log"

    logger = logging.getLogger("ibeatles_CLI")
    logger.setLevel(log_level)

    # Avoid adding handlers if they are already configured
    if not logger.hasHandlers():
        # Create file handler and stream handler
        file_handler = logging.FileHandler(log_file)
        stream_handler = logging.StreamHandler()

        # Set level for each handler
        file_handler.setLevel(log_level)
        stream_handler.setLevel(log_level)

        # Define formatter and add it to handlers
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        file_handler.setFormatter(formatter)
        stream_handler.setFormatter(formatter)

        # Add handlers to the logger
        logger.addHandler(file_handler)
        logger.addHandler(stream_handler)

    return logger


def load_config(config_path: Path) -> IBeatlesUserConfig:
    """
    Load and parse the configuration file.

    Parameters
    ----------
    config_path : Path
        Path to the JSON configuration file.

    Returns
    -------
    IBeatlesUserConfig
        Parsed configuration object.
    """
    logger = logging.getLogger("ibeatles_CLI")
    logger.info(f"Loading configuration: {config_path}")

    with open(config_path, "r") as f:
        config_data = json.load(f)
    return IBeatlesUserConfig(**config_data)


def load_data(config: IBeatlesUserConfig) -> Dict[str, Any]:
    """
    Load raw data, open beam, and spectra files.

    Parameters
    ----------
    config : IBeatlesUserConfig
        Parsed configuration object.

    Returns
    -------
    Dict[str, Any]
        Dictionary containing loaded data.
    """
    logger = logging.getLogger("ibeatles_CLI")
    logger.info("Loading data...")

    # Raw data is mandatory
    raw_data = load_data_from_folder(
        config.raw_data.raw_data_dir,
        file_extension=config.raw_data.extension,
    )
    # Open beam is optional
    if config.open_beam:
        open_beam = load_data_from_folder(
            config.open_beam.open_beam_data_dir,
            file_extension=config.open_beam.extension,
        )
    else:
        open_beam = None
    # Spectra file is needed, but specify the path in the configuration file is optional
    if config.spectra_file_path:
        spectra = load_time_spectra(
            config.spectra_file_path,
            config.analysis.distance_source_detector_in_m,
            config.analysis.detector_offset_in_us,
        )
    else:
        # try to load spectra file from the raw data directory
        spectra_file = get_time_spectra_filename(config.raw_data.raw_data_dir)
        if spectra_file:
            spectra = load_time_spectra(
                spectra_file,
                config.analysis.distance_source_detector_in_m,
                config.analysis.detector_offset_in_us,
            )
        else:
            raise ValueError("Spectra file not found")
    return {"raw_data": raw_data, "open_beam": open_beam, "spectra": spectra}


def perform_binning(data: Dict[str, Any], config: IBeatlesUserConfig, spectra_dict: dict) -> (Dict[str, Any], list):
    """
    Perform binning on the normalized data.

    Parameters
    ----------
    data : Dict[str, Any]
        Dictionary containing normalized data.
    config : IBeatlesUserConfig
        Parsed configuration object.
    spectra_dict:
        Dictionary containing time spectra data.

    Returns
    -------
    Dict[str, Any]
        Dictionary containing binning results.
    List[Tuple[slice]]
        List of bin coordinates.
    """
    logger = logging.getLogger("ibeatles_CLI")
    logger.info("Performing binning...")

    # Build binning coordinates
    bins = get_bin_coordinates(
        image_shape=data[0].shape,
        **config.analysis.pixel_binning.model_dump(),  # to dict for unpacking
    )
    # extract wavelength data from spectra dict
    # default unit is SI unit (meters)
    wavelengths_m = spectra_dict["lambda_array"]
    # execute binning
    bin_transmission = {}
    for i, bin_coord in enumerate(bins):
        wavelengths_bin, transmission_bin = get_bin_transmission(
            images=data,
            wavelengths=wavelengths_m,
            bin_coords=bin_coord,
            lambda_range=None,
        )
        bin_transmission[str(i)] = {
            "wavelengths": wavelengths_bin,
            "transmission": transmission_bin,
            "coordinates": bin_coord,
        }

    return bin_transmission, bins


def perform_fitting(bin_transmission_dict: Dict[str, Any], config: IBeatlesUserConfig) -> Dict[str, Any]:
    """
    Perform fitting on the normalized data.

    Parameters
    ----------
    bin_transmission_dict : Dict[str, Any]
        Dictionary containing binning results, from function perform_binning.
    config : IBeatlesUserConfig
        Parsed configuration object.

    Returns
    -------
    Dict[str, Any]
        Dictionary containing fitting results.
    """
    logger = logging.getLogger("ibeatles_CLI")
    logger.info("Performing fitting...")

    # step_0: prepare the lambda range
    lambda_min_angstrom = config.analysis.fitting.lambda_min * 1e10
    lambda_max_angstrom = config.analysis.fitting.lambda_max * 1e10
    lambda_range_angstrom = lambda_min_angstrom, lambda_max_angstrom
    # step_1: get the reference (zero strain) Bragg edge value
    lambda_0_angstrom = get_initial_bragg_edge_lambda(
        material_config=config.analysis.material,
        lambda_range=lambda_range_angstrom,
    )
    # step_2: setup the initial guess and bounds
    # NOTE: the only critical value here is the reference Bragg edge wavelength
    initial_parameters = {
        "a0": 0.1,
        "b0": 0.1,
        "a_hkl": 0.1,
        "b_hkl": 0.1,
        "bragg_edge_wavelength": lambda_0_angstrom,  # use the reference Bragg edge as the initial guess
        "sigma": 0.01,
        "tau": 0.01,
    }
    parameter_bounds = {
        "bragg_edge_wavelength": {
            "min": lambda_min_angstrom,
            "max": lambda_max_angstrom,
        },
        "sigma": {"min": 0.001, "max": 0.2},
        "tau": {"min": 0.001, "max": 0.2},
    }
    # step_3: fitting
    fit_results = {}  # (str(bin_id): lmfit.model.ModelResult)
    for key, value in bin_transmission_dict.items():
        wavelengths_angstrom = value["wavelengths"] * 1e10
        transmission = value["transmission"]
        # step_3.1: prepare the fitting range
        mask = (wavelengths_angstrom > lambda_min_angstrom) & (wavelengths_angstrom < lambda_max_angstrom)
        wavelengths_fitting_angstrom = wavelengths_angstrom[mask]
        transmission_fitting = transmission[mask]
        # step_3.2: fitting a smooth version first to get better initial guess
        # NOTE: eventually we will always get a fit for over-smoothed data, so we need to gradually increase the sigma
        #       although the quality of the initial guess decreases with the increase of sigma
        ratio = 0.10
        fit_success = False
        while not fit_success:
            sigma = int(len(transmission_fitting) * ratio)
            transmission_smooth = gaussian_filter1d(transmission_fitting, sigma=sigma)
            fit_result_smoothed = fit_bragg_edge_single_pass(
                wavelengths=wavelengths_fitting_angstrom,
                transmission=transmission_smooth,
                initial_parameters=initial_parameters,
                parameter_bounds=parameter_bounds,
            )
            if fit_result_smoothed is None:
                logger.info(f"Bin_{key}: Failed fitting with sigma = {sigma}, increase sigma and try again...")
                ratio += 0.02
                continue
            else:
                fit_success = True
        # step_3.3: fitting
        fit_result = fit_bragg_edge_single_pass(
            wavelengths=wavelengths_fitting_angstrom,
            transmission=transmission_fitting,
            initial_parameters=fit_result_smoothed.best_values,
            parameter_bounds=parameter_bounds,
        )
        fit_results[key] = fit_result

    return fit_results


def main(config_path: Union[Path, IBeatlesUserConfig], log_file: Optional[Path] = None) -> None:
    """
    Main function to run the iBeatles CLI application.

    Parameters
    ----------
    config_path : Path or IBeatlesUserConfig
        Path to the configuration file or the config object.
    log_file : Path, optional
        Path to the log file.

    Returns
    -------
    None
    """
    logger = setup_logging(log_file)

    try:
        # Load configuration
        if isinstance(config_path, IBeatlesUserConfig):
            config = config_path
        else:
            config = load_config(config_path)

        # Load data
        rst_dict = load_data(config)
        raw_data_dict = rst_dict["raw_data"]
        open_beam_dict = rst_dict["open_beam"]
        spectra_dict = rst_dict["spectra"]

        # Perform normalization
        normalized_data, output_path = normalize_data(
            sample_data=raw_data_dict["data"],
            ob_data=open_beam_dict["data"] if open_beam_dict else None,
            time_spectra=spectra_dict,
            config=config,
            output_folder=config.output["normalized_data_dir"],
        )
        logger.info(f"Normalized data saved to {output_path}.")

        # Binning
        binning_results, bins = perform_binning(
            data=normalized_data,
            config=config,
            spectra_dict=spectra_dict,
        )

        # Fitting
        logger.info("Performing fitting and strain mapping...")
        fitting_results = perform_fitting(
            bin_transmission_dict=binning_results,
            config=config,
        )
        # plot
        logger.info("Plotting fitting results grid...")
        lambda_min_angstrom = config.analysis.fitting.lambda_min * 1e10
        lambda_max_angstrom = config.analysis.fitting.lambda_max * 1e10
        lambda_range_angstrom = lambda_min_angstrom, lambda_max_angstrom
        lambda_0_angstrom = get_initial_bragg_edge_lambda(
            material_config=config.analysis.material,
            lambda_range=lambda_range_angstrom,
        )
        fig_fitting, _ = plot_fitting_results_grid(
            fit_results=fitting_results,
            bin_transmission=binning_results,
            reference_wavelength=lambda_0_angstrom,
        )
        # save figures
        logger.info("Saving fitting results...")
        output_path_strain = config.output.get("strain_results_dir", Path.cwd())
        # create the folder if not exist
        output_path_strain.mkdir(parents=True, exist_ok=True)
        # save fitting grid
        fn_fitting = generate_output_filename(
            input_folder=config.raw_data.raw_data_dir,
            analysis_type="fitting_grid",
            extension=config.analysis.strain_mapping.output_file_config.fitting_grid_format,
        )
        save_fitting_grid(
            figure=fig_fitting,
            output_path=output_path_strain / fn_fitting,
            config=config.analysis.strain_mapping.output_file_config,
        )
        # close the figure
        plt.close(fig_fitting)
        logger.info(f"Fitting results saved to {output_path_strain}.")

        # Calculate strain
        logger.info("Calculating strain mapping...")
        strain_results = calculate_strain_mapping(
            fit_results=fitting_results,
            d0=lambda_0_angstrom / 2.0,
            quality_threshold=config.analysis.strain_mapping.quality_threshold,
        )
        # plot
        logger.info("Plotting strain map overlay...")
        fig_strain, _ = plot_strain_map_overlay(
            strain_results=strain_results,
            bin_transmission=binning_results,
            integrated_image=np.sum(normalized_data, axis=0).T,
            colormap=config.analysis.strain_mapping.visualization.colormap,
            interpolation=config.analysis.strain_mapping.visualization.interpolation_method,
            alpha=config.analysis.strain_mapping.visualization.alpha,
        )
        # save figures
        logger.info("Saving strain map overlay...")
        fn_strain = generate_output_filename(
            input_folder=config.raw_data.raw_data_dir,
            analysis_type="strain_map",
            extension=config.analysis.strain_mapping.output_file_config.strain_map_format,
        )
        save_strain_map(
            figure=fig_strain,
            output_path=output_path_strain / fn_strain,
            config=config.analysis.strain_mapping.output_file_config,
        )
        # close the figure
        plt.close(fig_strain)
        logger.info(f"Strain map overlay saved to {output_path_strain}.")

        # Save analysis results
        logger.info("Saving analysis results...")
        output_path_analysis = config.output.get("analysis_results_dir", Path.cwd())
        output_path_analysis.mkdir(parents=True, exist_ok=True)
        # save analysis results
        fn_analysis = generate_output_filename(
            input_folder=config.raw_data.raw_data_dir,
            analysis_type="analysis_results",
            extension="csv",
        )
        # check if using custom material
        if config.analysis.material.custom_material:
            material_name = config.analysis.material.custom_material.name
        else:
            material_name = config.analysis.material.element
        # build the metadata header
        metadata = {
            "material_name": material_name,
            "d0": lambda_0_angstrom,
            "distance_source_detector": config.analysis.distance_source_detector_in_m,
            "detector_offset": config.analysis.detector_offset_in_us,
        }
        save_analysis_results(
            fit_results=fitting_results,
            bin_coordinates=bins,
            strain_results=strain_results,
            metadata=metadata,
            output_path=output_path_analysis / fn_analysis,
            config=config.analysis.strain_mapping.output_file_config,
        )
        logger.info(f"Analysis results saved to {output_path_strain}.")

        logger.info("iBeatles CLI application completed successfully.")
    except Exception as e:
        logger.error(f"An error occurred: {str(e)}")
        raise


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="iBeatles CLI Application")
    parser.add_argument("config", type=Path, help="Path to the configuration file")
    parser.add_argument("--log", type=Path, help="Path to the log file (optional)")
    args = parser.parse_args()

    main(args.config, args.log)
